#include "togl.h"
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>

#include "estypes.h"

#include "trackball.h"
#include "symmetrygroup.h"
// OTetrahedron
#include "ottile1.h"
#include "ottile2.h"
// STetraheron
#include "stetrahedron.h"
// DTetrahedron
#include "dtetrahedron.h"
// OOctahedron
#include "ootile1.h"
#include "ootile2.h"
#include "ootile3.h"
// OCube
#include "octile1.h"
#include "octile2.h"
#include "octile3.h"
// OIcosahedron
#include "oitile1.h"
#include "oitile2.h"
#include "oitile3.h"
// ODodecahedron
#include "odtile1.h"
#include "odtile2.h"
#include "odtile3.h"
// PentaDodecahedron
#include "pentadodecahedron.h"

#if defined(WIN32)
EXTERN int		TkConsoleInit(Tcl_Interp *interp);
#endif /* WIN32 */


// Called by Tk_Main() to let me init the modules (Togl) I will need.
int Tcl_AppInit(Tcl_Interp *pInterp);


enum DrawType 
{
  dtSPHERE,
  dtPOLYHEDRA,
  dtSTEREO,
  dtTRI_TILE,
  dtSOLID,
  dtSize
};

enum GridType 
{
  gtNO,
  gtFACE,
  gtTILE,
  gtSize
};

enum SymmetryGroupType 
{
  sgtOTetrahedron,
  sgtSTetrahedron,
  sgtDTetrahedron,
  sgtOOctahedron,
  sgtSOctahedron,
  sgtOCube,
  sgtSCube,
  sgtOIcosahedron,
  sgtSIcosahedron,
  sgtODodecahedron,
  sgtSDodecahedron,
  sgtPentaDodecahedron,
  sgtSize
};

static CHAR *gpcInputFile = NULL;

UINT_32 guiSpecialUI = 0;

static Togl *toglBall = NULL;
static UINT_32 iToglWidth=0;
static UINT_32 iToglHeight=0;
static Tcl_Interp *pTclInterp;
static BOOL gbGUICreated = FALSE;
BOOL gbZoomIn = FALSE;
GLuint GLuiSphere1_30;
GLuint GLuiSphere01_8;
GLuint GLuiSphere02_8;
GLuint GLuiSphere005_8;

static char gpcBuffer[200];
static char gpcBuffer2[200];
//WM
static char symmetryGroupType[100];

char *gpcTriangleCommand = NULL;
char **gppcTriangleArgs = new CHAR*[1];
static SymmetryGroup *pSG = NULL;
static TilePoint *gtptLastInnerPoint = NULL;
static DrawType dtDrawType = dtSPHERE;
GridType gtDrawGrid = gtFACE;
BOOL gbDrawNorthPole = FALSE;
BOOL gbPaintMode = FALSE;
BOOL gbDropperMode = FALSE;
Vector gvColor;
static BOOL gbDrawInnerPoints = FALSE;
static BOOL gbDrawHolePoint = FALSE;
static BOOL gbDeleteButtonDown = FALSE;
static BOOL gbMouseMotion = TRUE;
BOOL gbExtrudeInward = TRUE;
BOOL gbTileDetailed = TRUE;
BOOL gbCreateTriFiles = TRUE;

// Selection
static UINT_32 guiSelectedType  = SELECTED_NOTHING;
static UINT_32 guiSelectedIndex = SELECTED_NOTHING;
static UINT_32 guiSelectedHeightIndex = SELECTED_NOTHING;
UINT_32 guiCurrentSelectedInner = SELECTED_NOTHING;
static BOOL gbLeftButtonDown = FALSE;
static BOOL gbMiddleButtonDown = FALSE;
static BOOL gbRightButtonDown = FALSE;
static BOOL gbDeletingPoint = FALSE;
//WM
static UINT_32 guiColorIndex = 0;

// Ball Camera UI
static Matrix mBallRotate;    // Rotation component
static Matrix mBallInvRotate; // Inverse rotation
static Matrix mBallScale;
static Matrix mBallTranslate;

static Matrix mBallCamera;
static Vector vBallNormMouseLast;
static FLOAT  fBallScale = 1.0;

// track ball
static int iWidgetLastX = 0;
static int iWidgetLastY = 0;
static int iCenterX = 0;       // keep track of the sphere center
static int iCenterY = 0;       
static int iMouseX, iMouseY;    // last x, y in viewport coords
static float curquat[4];      // current rotation quaternion 
static float lastquat[4];     // last rotation quaternion 
static float invlastquat[4];  // inverse of last rotation quaternion 
static int spinning;          // spin sphere
static int time;              // time between mouse motions 
static float velocity;        // speed of current mouse motion

//////////////////////////////////////////////////////////////////////////////////
// Functions

//WM convert integer to hexadecimal and return it as a string
VOID IntConvertString (int num, CHAR* rString) 
{
  int result, quotient, i;
  result = num/16;
  quotient = num%16;
  
  for( i = 0; i < 2; i++)
    {
      switch(result) 
	{
	   case 0: rString[i] = '0';
	     break;
	   case 1: rString[i] = '1';
	     break;
	   case 2: rString[i] = '2';
	     break;
	   case 3: rString[i] = '3';
	     break;
	   case 4: rString[i] = '4';
	     break;
	   case 5: rString[i] = '5';
	     break;
	   case 6: rString[i] = '6';
	     break;
	   case 7: rString[i] = '7';
	     break;
	   case 8: rString[i] = '8';
	     break;
	   case 9: rString[i] = '9';
	     break;
	   case 10: rString[i] = 'a';
	     break;
	   case 11: rString[i] = 'b';
	     break;
	   case 12: rString[i] = 'c';
	     break;
	   case 13: rString[i] = 'd';
	     break;
	   case 14: rString[i] = 'e';
	     break;
	   case 15: rString[i] = 'f';
	}
      result = quotient;
    }
}

VOID DrawSphere(FLOAT fRadius, UINT_32 uiZSlices, UINT_32 uiThetaSlices )
{
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fZMin, fZMax, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Vector vNormal;
  Point pppt[2][2];

  fZMin = -1;
  fZMax =  1;
  fPhiMin = DEGTORAD(-90);
  fPhiMax = DEGTORAD(90);
  fThetaMin = 0.0;
  fThetaMax = DEGTORAD(360);
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
  for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
    {
      fZ1 = fRadius*(FLOAT)sin(fPhi1);
      fR1 = fRadius*(FLOAT)cos(fPhi1);
      fZ0 = fRadius*(FLOAT)sin(fPhi0);
      fR0 = fRadius*(FLOAT)cos(fPhi0);
      
      glBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;
	  
	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	pppt[1][1].Set(fR1*(FLOAT)cos(fTheta+fThetaDelta), 
		       fR1*(FLOAT)sin(fTheta+fThetaDelta), fZ1);
	  
	vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	vNormal.Normalize();
	  
	glNormal3fv(vNormal.Get());
	glVertex3fv(pppt[1][0].Get());
	glVertex3fv(pppt[0][0].Get());
	
	fTheta += fThetaDelta;	
	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	    
	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    glNormal3fv(vNormal.Get());
	    glVertex3fv(pppt[1][1].Get());
	    
	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    glNormal3fv(vNormal.Get());
	    glVertex3fv(pppt[0][1].Get());
	      
	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;
	  }
      }
      glEnd(); 

      fPhi0 = fPhi1;
      fPhi1 += fPhiDelta;
    }
}

VOID Init()
{
  GLuiSphere1_30  = glGenLists(1);
  GLuiSphere01_8 = glGenLists(1);
  GLuiSphere02_8 = glGenLists(1);
  GLuiSphere005_8 = glGenLists(1);

  glNewList( GLuiSphere1_30, GL_COMPILE );
  {
    DrawSphere(0.9999, 30, 30);
  }
  glEndList();

  glNewList( GLuiSphere01_8, GL_COMPILE );
  {
    DrawSphere(0.01, 8, 8);
  }
  glEndList();

  glNewList( GLuiSphere02_8, GL_COMPILE );
  {
    DrawSphere(0.02, 8, 8);
  }
  glEndList();

  glNewList( GLuiSphere005_8, GL_COMPILE );
  {
    DrawSphere(0.005, 8, 8);
  }
  glEndList();

  iToglWidth = Togl_Width(toglBall);
  iToglHeight = Togl_Height(toglBall);

  gvColor.Set(132.0/255.0, 193.0/255.0, 214.0/255.0); // = tcl lightblue

  // Get Initial Directory
  INT_32 iCode;

  sprintf(gpcBuffer, "pwd");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  ASSERT( pTclInterp->result != NULL );

  gpcTriangleCommand = new CHAR[strlen(pTclInterp->result) + 
			       strlen("/triangle -pe es_tri.poly") + 1];
  sprintf(gpcTriangleCommand, "%s/triangle -pe es_tri.poly", pTclInterp->result);


  gppcTriangleArgs[0] = new CHAR[strlen(pTclInterp->result) + strlen("/triangle") + 1];
  gppcTriangleArgs[1] = new CHAR[strlen("-pe") + 1];
  gppcTriangleArgs[2] = new CHAR[strlen("es_tri.poly") + 1];

  sprintf(gppcTriangleArgs[0], "%s/triangle", pTclInterp->result);
  sprintf(gppcTriangleArgs[1], "-pe");
  sprintf(gppcTriangleArgs[2], "es_tri.poly");
}

VOID SetStatusLabel(CHAR *pcLabel)
{
  int iCode;

  sprintf(gpcBuffer, "SetStatusLabel \"%s\"", pcLabel);
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);

  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }
}

FLOAT Norm2ModelSpace(FLOAT fNorm)
{
  return (5.0/2.0)*fNorm/fBallScale;
}

FLOAT ModelSpace2Norm(FLOAT fNorm)
{
  return (2.0/5.0)*fBallScale/fNorm;
}

BOOL SphereIntersection( Point &pt, struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY)
{
  FLOAT f, fX, fY, fNormX, fNormY, fNormCenterX, fNormCenterY;
  INT_32 iWidth, iHeight;

  iWidth = iToglWidth;
  iHeight = iToglHeight;

  // Normalize widget coords
  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;
  fNormCenterX = (float)(iCenterX + (iToglWidth/2.0)) / iWidth ;  
  fNormCenterY = (float)(iHeight - (iCenterY + iToglHeight/2.0)) / iHeight; 

  // Calculate x and y in viewport -> model space distances
  fX = Norm2ModelSpace( fNormX-fNormCenterX );
  fY = Norm2ModelSpace( fNormY-fNormCenterY );
  pt.Set( fX, fY, 0.0 );

  // Check if the cursor is inside the circle
  f = 1.0 - pt[X]*pt[X] - pt[Y]*pt[Y];
  if( f >= 0 )
    {
      pt[Z] = sqrt(f);
      pt[W] = 1.0;
      pt = mBallInvRotate * pt;
      return TRUE;
    }

  return FALSE;
}

FLOAT CalculateHeight( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  /*
  FLOAT f, fX, fY, fNormX, fNormY, m, a, b, D;
  Point ptMouse, p1, p2, pNorm;
  INT_32 iWidth, iHeight;
  Vector vMouse, vNormal;

  iWidth = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  // Normalize widget coords
  fNormX = (float)iWidgetLastX / iWidth;
  fNormY = (float)(iHeight - iWidgetLastY) / iHeight;
  ptMouse.Set( fNormX, fNormY, 0 );

  pSG->GetInnerPointNormalPoints(p1, p2);
  p1 = ModelSpace2Norm(p1);
  p2 = ModelSpace2Norm(p2);

  if( (p1[X] - p2[X]) == 0 ) // x = c;
    {
    }
  else if( (p1[Y] - p2[Y]) == 0) // y = c;
    {
    }
  else
    {
      vNormal = p2 - p1;
      vNormal.Normalize();

      m = (p1[Y] - p2[Y])/(p1[X] - p2[X]);
      a = m;
      b = -1;
      c = p1[Y] - m*p1[X];
      
      D = (a*fNormX + b*fNormY + c)/sqrt(a*a + b*b);

      pt = ptMouse - D*vNormal;

      t = (pt[X] - p2[X])/(p1[X] - p2[X]);
      pt[Z] = p1[Z]*t + p2[Z]*(1.0 - t);

  
    }

    return f;
    */

  FLOAT f, fX1, fX2, fY1, fY2, fNormX1, fNormY1, fNormX2, fNormY2;
  INT_32 iWidth, iHeight;
  Vector vMouse, vNormal;

  iWidth = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  // Normalize widget coords
  fNormX1 = (float)iWidgetLastX / iWidth;
  fNormY1 = (float)(iHeight - iWidgetLastY) / iHeight;
  fNormX2 = (float)iWidgetX / iWidth;
  fNormY2 = (float)(iHeight - iWidgetY) / iHeight;

  // Calculate x and y in viewport -> model space distances
  fX1 = Norm2ModelSpace( fNormX1 );
  fY1 = Norm2ModelSpace( fNormY1 );
  fX2 = Norm2ModelSpace( fNormX2 );
  fY2 = Norm2ModelSpace( fNormY2 );
  vMouse.Set( fX2-fX1, fY2-fY1, 0.0 );

  // Crystal Balled World
  vNormal = mBallRotate * pSG->GetPointNormalVector(guiSelectedIndex);
  f = vMouse.Dot(vNormal);

  return f;
}

CHAR *GetFileDirPath (CHAR *pcFullPath)
{
  INT_32 iLastslash;
  iLastslash = -1;
  CHAR *temp;

  // Grab the directory name out of this fully-qualified filename
  // Is there a premade function to do this?  maybe...
  // 
  for (UINT_32 i=0; i< strlen (pcFullPath); i++)
  {
	  if (pcFullPath[i] == '/')
	  {
		  iLastslash = i;
	  }
  }
  if (iLastslash >= 0)
  {
	  // if not in current directory, CD there to be in the right
	  // directory.
	  //
	  temp = new char[iLastslash + 3];
	  strncpy ( temp, pcFullPath, iLastslash+1);
	  temp[iLastslash+1] = '\0';
  }
  else
  {
	  temp = new char[1];
	  temp[0] = '\0';
  }

  return temp;
}

//////////////////////////////////////////////////////////////////////////////////
// Tcl Callbacks

CHAR* ExecuteTCLCommand(CHAR *pcCommand)
{
  int iCode;

  sprintf(gpcBuffer, "%s", pcCommand);
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);

  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }

  return pTclInterp->result;
}

VOID ResetCamera()
{
  mBallRotate.SetIdentity();
  mBallInvRotate.SetIdentity();
  mBallScale.SetIdentity();
  mBallTranslate.SetIdentity();
  
  mBallCamera.SetIdentity();
  fBallScale = 1.0;
  
  iCenterX = 0;
  iCenterY = 0;
}

VOID ResetInterface()
{
  int iCode;

  gbDrawInnerPoints = FALSE;
  sprintf(gpcBuffer, "resetInnerPointsButton");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }
  
  gbDrawHolePoint = FALSE;
  sprintf(gpcBuffer, "resetHolePointButton");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }
  
  gbDeleteButtonDown = FALSE;
  sprintf(gpcBuffer, "resetDeletePointButton");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }

  if( pSG != NULL )
    {
      pSG->SetThickness(0.02);
      pSG->SetRadius(1.0);
    }
  sprintf(gpcBuffer, "set thickness 0.02; set radius 1.0");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }

  sprintf(gpcBuffer, "set tiles 1");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }
}

int CBHandleGUICreated( ClientData clientData, Tcl_Interp *pInterp, 
			int argc, char **argv)
{
  ASSERT( argc == 1 );

  gbGUICreated = TRUE;

  return TCL_OK;
}

int CBHandleHolePoint( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  gbDrawHolePoint = !gbDrawHolePoint;
  return TCL_OK;
}

int CBHandleDeleteDown( ClientData clientData, Tcl_Interp *pInterp, 
			int argc, char **argv)
{
  ASSERT( argc == 1 );
  
  gbDeleteButtonDown = !gbDeleteButtonDown;
  return TCL_OK;
}

int CBHandleDeselectInner( ClientData clientData, Tcl_Interp *pInterp, 
			   int argc, char **argv)
{
  ASSERT( argc == 1 );
  
  if( pSG != NULL )
    {
      pSG->DeselectAllHeightPoints();
      Togl_PostRedisplay( toglBall );
    }

  return TCL_OK;
}

int CBHandleInnerPoints( ClientData clientData, Tcl_Interp *pInterp, 
			 int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      if( gbDrawInnerPoints )
	{
	  gbDrawInnerPoints = FALSE;

	  if( gtptLastInnerPoint != NULL )
	    {
	      gtptLastInnerPoint->SetConnected(FALSE);
	    }
	  
	  guiSelectedType  = SELECTED_NOTHING;
	  guiSelectedIndex = SELECTED_NOTHING;
	}
      else
	{
	  gbDrawInnerPoints = TRUE;
	}
    }
  return TCL_OK;
}

int CBHandleSetSymmetryGroup( ClientData clientData, Tcl_Interp *pInterp, 
			      int argc, char **argv)
{
  ASSERT( argc == 3 );

  CHAR hexString[7];

  int iSymmetryGroup, iSymmetryGroupTile, iCode;

  iSymmetryGroup     = atoi( argv[1] );
  iSymmetryGroupTile = atoi( argv[2] );

  if( pSG != NULL )
    {
      delete pSG;
      pSG = NULL;
      strcpy(symmetryGroupType, "");

      hexString[0] = '#';
      IntConvertString(128, &(hexString[1]));
      IntConvertString(128, &(hexString[3]));
      IntConvertString(128, &(hexString[5]));
    
      sprintf(gpcBuffer2, "resetAllButton %.7s", hexString);
      ExecuteTCLCommand(gpcBuffer2);
    }

  if( gbGUICreated )
    {
      ResetInterface();

      if( (dtDrawType != dtSPHERE) && (dtDrawType != dtPOLYHEDRA) )
	{
	  dtDrawType = dtSPHERE;

	  sprintf(gpcBuffer, "set draw 0");
	  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
	  if ( iCode != TCL_OK )
	    {
	      printf( "Error: %s:\n", gpcBuffer );
	      printf("Error Message: %s \n", pTclInterp->result);
	    }

	  ResetCamera();
	}
    }

  switch( iSymmetryGroup )
    {
    case sgtOTetrahedron:
      {
	// OTetrahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OTTile1();
	      strcpy(symmetryGroupType, "otetrahedron tile 1");
	    }
	  break;
	  case(2):
	    {
	      pSG = new OTTile2(); 
	      strcpy(symmetryGroupType, "otetrahedron tile 2");
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtSTetrahedron:
      {
	// STetrahedron
	pSG = new STetrahedron(); 
	strcpy(symmetryGroupType, "stetrahedron tile 0");
      }
    break;
    case sgtDTetrahedron:
      {
	// DTetrahedron
	pSG = new DTetrahedron(); 
	strcpy(symmetryGroupType, "dtetrahedron tile 0");
      }
    break;
    case sgtOOctahedron:
      {
	// OOctahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OOTile1(); 
	      strcpy(symmetryGroupType, "ooctahedron tile 1");
	    }
	  break;
	  case(2):
	    {
	      pSG = new OOTile2(); 
	      strcpy(symmetryGroupType, "ooctahedron tile 2");
	    }
	  break;
	  case(3):
	    {
	      pSG = new OOTile3(); 
	      strcpy(symmetryGroupType, "ooctahedron tile 3");
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtOCube:
      {
	// OCube
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OCTile1(); 
	      strcpy(symmetryGroupType, "ocube tile 1");
	    }
	  break;
	  case(2):
	    {
	      pSG = new OCTile2(); 
	      strcpy(symmetryGroupType, "ocube tile 2");
	    }
	  break;
	  case(3):
	    {
	      pSG = new OCTile3(); 
	      strcpy(symmetryGroupType, "ocube tile 3");
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtOIcosahedron:
      {
	// OCube
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OITile1(); 
	      strcpy(symmetryGroupType, "oicosahedron tile 1");
	    }
	  break;
	  case(2):
	    {
	      pSG = new OITile2(); 
	      strcpy(symmetryGroupType, "oicosahedron tile 2");
	    }
	  break;
	  case(3):
	    {
	      pSG = new OITile3(); 
	      strcpy(symmetryGroupType, "oicosahedron tile 3");
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtODodecahedron:
      {
	// ODodecahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new ODTile1(); 
	      strcpy(symmetryGroupType, "ododecahedron tile 1");
	    }
	  break;
	  case(2):
	    {
	      pSG = new ODTile2(); 
	      strcpy(symmetryGroupType, "ododecahedron tile 2");
	    }
	  break;
	  case(3):
	    {
	      pSG = new ODTile3(); 
	      strcpy(symmetryGroupType, "ododecahedron tile 3");
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtPentaDodecahedron:
      {
	// PentaDodecahedron
	pSG = new PentaDodecahedron(); 
	strcpy(symmetryGroupType, "pentadodecahedron tile 0");
      }
    break;
    default:
      {
	pSG = NULL;
	ASSERT( FALSE );      
      }
    break;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleDrawNorthPole( ClientData clientData, Tcl_Interp *pInterp, 
			   int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      gbDrawNorthPole = !gbDrawNorthPole;
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetDraw( ClientData clientData, Tcl_Interp *pInterp, 
		     int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      dtDrawType = (DrawType)(atoi( argv[1] ));
      
      if((dtDrawType == dtSTEREO) || (dtDrawType == dtTRI_TILE) || (dtDrawType == dtSOLID))
	{
	  pSG->TriangulateTile();
	}
      
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetPaintMode( ClientData clientData, Tcl_Interp *pInterp, 
			  int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( (dtDrawType != dtSOLID) || gbPaintMode )
    {
      gbPaintMode = FALSE;
      ExecuteTCLCommand("paintModeOff");
      return TCL_OK;
    }

  if( pSG != NULL ) 
    {
      gbPaintMode = TRUE;
      gbDropperMode = FALSE;
    }
  return TCL_OK;
}

//WM
int CBHandleSetDropperMode( ClientData clientData, Tcl_Interp *pInterp, 
			    int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( (dtDrawType != dtSOLID) || gbDropperMode )
    {
      gbDropperMode = FALSE;
      ExecuteTCLCommand("dropperModeOff");
      return TCL_OK;
    }

  if( pSG != NULL ) 
    {
      gbDropperMode = TRUE;
      gbPaintMode = FALSE;
    }
  return TCL_OK;
}

int CBHandleSetPaintColor( ClientData clientData, Tcl_Interp *pInterp, 
			   int argc, char **argv)
{
  ASSERT( argc == 4 );

  FLOAT r, g, b;

  r = atoi( argv[1] ) / 65536.0;
  g = atoi( argv[2] ) / 65536.0;
  b = atoi( argv[3] ) / 65536.0;

  gvColor.Set(r, g, b);

  return TCL_OK;
}

int CBHandleIncrUI( ClientData clientData, Tcl_Interp *pInterp, 
		    int argc, char **argv)
{
  ASSERT( argc == 1 );

  // I3D Video
  //printf( "IncrUI: %lu\n", guiSpecialUI );

  guiSpecialUI++;
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleDecrUI( ClientData clientData, Tcl_Interp *pInterp, 
		    int argc, char **argv)
{
  ASSERT( argc == 1 );

  // I3D Video
  if( guiSpecialUI > 0 )
    {
      guiSpecialUI--;     
      Togl_PostRedisplay( toglBall );
    }

  return TCL_OK;
}

int CBHandleSetRadius( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      FLOAT fRadius;
      
      fRadius = atof( argv[1] );
      pSG->SetRadius(fRadius);
      
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetThickness( ClientData clientData, Tcl_Interp *pInterp, 
			  int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      FLOAT fThickness;
      
      fThickness = atof( argv[1] );
      pSG->SetThickness(fThickness);
      
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetExtrude( ClientData clientData, Tcl_Interp *pInterp, 
			int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      UINT_32 uiInward;
      
      uiInward = atoi( argv[1] );
      gbExtrudeInward = uiInward;

      pSG->SetTileSolidModified(TRUE);

      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetShell( ClientData clientData, Tcl_Interp *pInterp, 
		     int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      UINT_32 uiDetailed;
      
      uiDetailed = atoi( argv[1] );
      gbTileDetailed = uiDetailed;

      pSG->SetTileSolidModified(TRUE);
      
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleSetSolidTiles( ClientData clientData, Tcl_Interp *pInterp, 
			   int argc, char **argv)
{
  ASSERT( argc == 2 );

  if( pSG != NULL )
    {
      BOOL bAll;

      bAll = atoi( argv[1] );

      pSG->SetAllSolidTiles(bAll);
      Togl_PostRedisplay( toglBall );
    }
  return TCL_OK;
}

int CBHandleTriFiles( ClientData clientData, Tcl_Interp *pInterp, 
		      int argc, char **argv)
{
  ASSERT( argc == 1 );

  gbCreateTriFiles = !gbCreateTriFiles;
  
  if( gbCreateTriFiles )
    printf( "Creating triangle files\n");    
  else
    printf( "Not creating triangle files\n");

  return TCL_OK;
}

int CBHandleDrawGrid( ClientData clientData, Tcl_Interp *pInterp, 
		      int argc, char **argv)
{
  ASSERT( argc == 2 );

  gtDrawGrid = (GridType)(atoi( argv[1] ));  
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleTriangulate( ClientData clientData, Tcl_Interp *pInterp, 
			 int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      pSG->TriangulateTile();
      Togl_PostRedisplay( toglBall );
    }

  return TCL_OK;
}

int CBHandleZoom( ClientData clientData, Tcl_Interp *pInterp, 
		  int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      gbZoomIn = !gbZoomIn;

      pSG->SetTileBasReliefModified(TRUE);

      Togl_PostRedisplay( toglBall );
    }

  return TCL_OK;
}

int CBHandleOutputSTL( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .stl -filetypes {{\"STL Files\" {.stl}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->TriangulateTile();
      pSG->OutputSTL(fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleOutputVRML( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .wrl -filetypes {{\"VRML Files\" {.wrl}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->TriangulateTile();
      pSG->OutputVRML(fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//WM
VOID UpdateButtonColor (UINT_32 uiSelectedIndex, FLOAT r, FLOAT g, FLOAT b) 
{
  CHAR hexString[7];
  int iR, iG, iB;

  hexString[0] = '#';
  iR = (int)(r*255);
  iG = (int)(g*255);
  iB = (int)(b*255);
  IntConvertString(iR, &(hexString[1]));
  IntConvertString(iG, &(hexString[3]));
  IntConvertString(iB, &(hexString[5]));
  //printf ("R = %f G = %f B = %f     ", r, g, b);
  //printf ("iR = %d iG = %d iB = %d     ", iR, iG, iB);
  //printf ("hexString = %.7s\n", hexString);
  
  sprintf(gpcBuffer2, "setColorButton %ld %.7s", uiSelectedIndex, hexString);
  ExecuteTCLCommand(gpcBuffer2);
}


int CBHandleOutputES( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;

      //WM
      FILE *fpColorIndexFile, *fpPaletteFile;
      CHAR *pcColorIndexFile, *pcPaletteFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .es -filetypes {{\"ES Files\" {.es}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;
      
      //Creating the paletteFile
      pcPaletteFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 2];
      ASSERT ( pcPaletteFile != NULL );
      strncpy(pcPaletteFile, Tcl_GetStringResult(pInterp), 
	      strlen(Tcl_GetStringResult(pInterp))-2);
      strcpy(&(pcPaletteFile[strlen(Tcl_GetStringResult(pInterp))-2]), "pes");
      fpPaletteFile = NULL;

      //Creating the colorIndexFile
      pcColorIndexFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 2];
      ASSERT ( pcColorIndexFile != NULL );
      strncpy(pcColorIndexFile, Tcl_GetStringResult(pInterp), 
	      strlen(Tcl_GetStringResult(pInterp))-2);
      strcpy(&(pcColorIndexFile[strlen(Tcl_GetStringResult(pInterp))-2]), "ces");
      fpPaletteFile = NULL;

      //Checking if we successfully open the 3 files
      while ( fpOutputFile == NULL ||
	      fpPaletteFile == NULL ||
	      fpColorIndexFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  fpPaletteFile = fopen(pcPaletteFile, "w");
	  fpColorIndexFile = fopen(pcColorIndexFile, "w");
	  
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }

	  if ( fpPaletteFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcPaletteFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	  
	  if ( fpColorIndexFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcColorIndexFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->OutputES(fpOutputFile, pcPaletteFile, pcColorIndexFile);
      pSG->OutputPES(fpPaletteFile);
      pSG->OutputCES(fpColorIndexFile, symmetryGroupType);

    cancel:
      if ( pcOutputFile != NULL)
	delete []pcOutputFile;
      if ( pcPaletteFile != NULL)
	delete []pcPaletteFile;
      if ( pcColorIndexFile != NULL)
	delete []pcColorIndexFile;
      
      if (( fpOutputFile != stdout && fpOutputFile != NULL ) ||
	  ( fpPaletteFile != stdout && fpPaletteFile != NULL ) ||
	  ( fpColorIndexFile != stdout && fpColorIndexFile != NULL ))
	{
	  ASSERT( fpOutputFile != NULL );
	  ASSERT( fpPaletteFile != NULL );
	  ASSERT( fpColorIndexFile != NULL );
	  fclose(fpOutputFile);
	  fclose(fpPaletteFile);
	  fclose(fpColorIndexFile);
	}
      
      fpOutputFile = NULL;
      fpPaletteFile = NULL;
      fpColorIndexFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//WM Handle Output PES
int CBHandleOutputPES( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;

      CHAR *pcPaletteFile;
      FILE *fpPaletteFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .pes -filetypes {{\"PES Files\" {.pes}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}

      //Creating the paletteFile
      pcPaletteFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT ( pcPaletteFile != NULL );
      strcpy(pcPaletteFile, Tcl_GetStringResult(pInterp));
      fpPaletteFile = NULL;
      
      while ( fpPaletteFile == NULL )
	{
	  fpPaletteFile = fopen(pcPaletteFile, "w");

	  //WM need to check that fpColorIndexFile and fpPaletteFile is not NULL
	  
	  if ( fpPaletteFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcPaletteFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->OutputPES(fpPaletteFile);

    cancel:
      if ( pcPaletteFile != NULL )
	{
	  delete []pcPaletteFile;
	}
      
      if ( fpPaletteFile != stdout && fpPaletteFile != NULL )
	{
	  ASSERT( fpPaletteFile != NULL );
	  fclose(fpPaletteFile);
	}
      
      fpPaletteFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//WM Handle Output CES
int CBHandleOutputCES( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;

      CHAR *pcColorIndexFile;
      FILE *fpColorIndexFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .ces -filetypes {{\"CES Files\" {.ces}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}

      //Creating the colorIndexFile
      pcColorIndexFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT ( pcColorIndexFile != NULL );
      strcpy(pcColorIndexFile, Tcl_GetStringResult(pInterp));
      fpColorIndexFile = NULL;
      
      while ( fpColorIndexFile == NULL )
	{
	  fpColorIndexFile = fopen(pcColorIndexFile, "w");
 
	  if ( fpColorIndexFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcColorIndexFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->OutputCES(fpColorIndexFile, symmetryGroupType);

    cancel:
      if ( pcColorIndexFile != NULL )
	{
	  delete []pcColorIndexFile;
	}
      
      if ( fpColorIndexFile != stdout && fpColorIndexFile != NULL )
	{
	  ASSERT( fpColorIndexFile != NULL );
	  fclose(fpColorIndexFile);
	}
      
      fpColorIndexFile = NULL;
    }

  return TCL_OK;
}


VOID ResetSymmetryGroup()
{
  INT_32 iCode;
  CHAR hexString[7];

  // Set menu to unselected
  sprintf(gpcBuffer, "set sym $eSize");
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }

  if( pSG != NULL )
    {
      delete pSG;
      pSG = NULL;
      strcpy(symmetryGroupType, "");

      hexString[0] = '#';
      IntConvertString(128, &(hexString[1]));
      IntConvertString(128, &(hexString[3]));
      IntConvertString(128, &(hexString[5]));
    
      sprintf(gpcBuffer2, "resetAllButton %.7s", hexString);
      ExecuteTCLCommand(gpcBuffer2);
    }
}

VOID ReadSymmetryGroup(FILE *file, CHAR *pcFile)
{
  INT_32 iCode;
  CHAR pcSymmetryGroup[30], pcFoo[30];
  UINT_32 uiTile;

  ResetSymmetryGroup();

  fscanf(file, "%s %s %ld", pcSymmetryGroup, pcFoo, &uiTile); 

  if( strcmp(pcSymmetryGroup, "otetrahedron") == 0 )
    {
      if( uiTile == 1 )
	{
	  pSG = new OTTile1();
	  strcpy(symmetryGroupType, "otetrahedron tile 1");
	}
      else if( uiTile == 2)
	{
	  pSG = new OTTile2(); 
	  strcpy(symmetryGroupType, "otetrahedron tile 2");
	}
      else
	{
	  printf( "ES file \"%s\" has an invalid symmetry group tile \"%ld\".\n", 
		  pcFile, uiTile);
	  return;
	}

      sprintf(gpcBuffer, "set sym $eOTetra");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "dtetrahedron") == 0 )
    {
      pSG = new DTetrahedron(); 
      strcpy(symmetryGroupType, "dtetrahedron tile 0");
      
      sprintf(gpcBuffer, "set sym $eDTetra");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "stetrahedron") == 0 )
    {
      pSG = new STetrahedron(); 
      strcpy(symmetryGroupType, "stetrahedron tile 0");

      sprintf(gpcBuffer, "set sym $eSTetra");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "ooctahedron") == 0 )
    {
      if( uiTile == 1 )
	{
	  pSG = new OOTile1(); 
	  strcpy(symmetryGroupType, "ooctahedron tile 1");
	}
      else if( uiTile == 2 )
	{
	  pSG = new OOTile2(); 
	  strcpy(symmetryGroupType, "ooctahedron tile 2");
	}
      else if( uiTile == 3 )
	{
	  pSG = new OOTile3(); 
	  strcpy(symmetryGroupType, "ooctahedron tile 3");
	}
      else
	{
	  printf( "ES file \"%s\" has an invalid symmetry group tile \"%ld\".\n", 
		  pcFile, uiTile);
	  return;
	}

      sprintf(gpcBuffer, "set sym $eOOcta");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "ocube") == 0 )
    {
      if( uiTile == 1 )
	{
	  pSG = new OCTile1(); 
	  strcpy(symmetryGroupType, "ocube tile 1");
	}
      else if( uiTile == 2 )
	{
	  pSG = new OCTile2(); 
	  strcpy(symmetryGroupType, "ocube tile 2");
	}
      else if( uiTile == 3 )
	{
	  pSG = new OCTile3(); 
	  strcpy(symmetryGroupType, "ocube tile 3");
	}
      else
	{
	  printf( "ES file \"%s\" has an invalid symmetry group tile \"%ld\".\n", 
		  pcFile, uiTile);
	  return;
	}

      sprintf(gpcBuffer, "set sym $eOCube");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "oicosahedron") == 0 )
    {
      if( uiTile == 1 )
	{
	  pSG = new OITile1(); 
	  strcpy(symmetryGroupType, "oicosahedron tile 1");
	}
      else if( uiTile == 2 )
	{
	  pSG = new OITile2(); 
	  strcpy(symmetryGroupType, "oicosahedron tile 2");
	}
      else if( uiTile == 3 )
	{
	  pSG = new OITile3(); 
	  strcpy(symmetryGroupType, "oicosahedron tile 3");
	}
      else
	{
	  printf( "ES file \"%s\" has an invalid symmetry group tile \"%ld\".\n", 
		  pcFile, uiTile);
	  return;
	}

      sprintf(gpcBuffer, "set sym $eOIcosa");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "ododecahedron") == 0 )
    {
      if( uiTile == 1 )
	{
	  pSG = new ODTile1(); 
	  strcpy(symmetryGroupType, "ododecahedron tile 1");
	}
      else if( uiTile == 2 )
	{
	  pSG = new ODTile2(); 
	  strcpy(symmetryGroupType, "ododecahedron tile 2");
	}
      else if( uiTile == 3 )
	{
	  pSG = new ODTile3(); 
	  strcpy(symmetryGroupType, "ododecahedron tile 3");
	}
      else
	{
	  printf( "ES file \"%s\" has an invalid symmetry group tile \"%ld\".\n", 
		  pcFile, uiTile);
	  return;
	}

      sprintf(gpcBuffer, "set sym $eODodeca");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else if( strcmp(pcSymmetryGroup, "pentadodecahedron") == 0 )
    {
      pSG = new PentaDodecahedron(); 
      strcpy(symmetryGroupType, "pentadodecahedron tile 0");

      sprintf(gpcBuffer, "set sym $ePDodeca");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
  else
    {
      printf( "ES file \"%s\" has an invalid symmetry group.\n", pcFile);
    }
}

VOID ReadSolid(FILE *file, CHAR *pcFile)
{
  ASSERT( pSG != NULL );

  INT_32 iCode;
  UINT_32 uiExtrude, uiShell;
  FLOAT fRadius, fThickness;
  CHAR pcFoo[30];

  fscanf(file, "%s %f %s %f %s %ld %s %ld", 
	 pcFoo, &fRadius, pcFoo, &fThickness,
	 pcFoo, &uiExtrude, pcFoo, &uiShell);
	
  if( (fRadius < 0) || !( (0.0 < fThickness) && (fThickness < 1.0) ) )
    {
      ResetSymmetryGroup();
      printf( "ES file \"%s\" has an invalid radius or thickness.\n", pcFile);
    }
  else
    {
      pSG->SetRadius(fRadius);
      pSG->SetThickness(fThickness);

      gbExtrudeInward = uiExtrude;
      gbTileDetailed = uiShell;

      sprintf(gpcBuffer, "set radius %f", fRadius);
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}

      sprintf(gpcBuffer, "set thickness %f", fThickness);
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}

      sprintf(gpcBuffer, "set extrude %ld", uiExtrude);
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}

      sprintf(gpcBuffer, "set shell %ld", uiShell);
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
    }
}

VOID ReadSphereOutlinePoints( FILE *file, CHAR *pcFile )
{
  ASSERT( pSG != NULL );

  CHAR pcFoo[30];
  FLOAT f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15;
  FLOAT fX, fY, fZ;
  UINT_32 i, uiID, uiIDPartner, uiType, uiTilePoint, uiSphereOutlinePoints;
  MATRIX4D mm;
  Matrix m;
  Point pt;

  pSG->ClearSphereOutlinePoints();

  fscanf(file, "%s %ld", pcFoo, &uiSphereOutlinePoints);

  for( i = 0; i < uiSphereOutlinePoints; i++ )
    {
      // Point location and type
      fscanf(file, "%s %f %f %f %s %ld", pcFoo, &fX, &fY, &fZ, pcFoo, &uiType);
      pt.Set( fX, fY, fZ );
      uiTilePoint = 0;
      m.SetIdentity();
      
      if( uiType == 0 ) // Tile Point
	{
	  // TilePoint #
	  fscanf(file, "%s %ld", pcFoo, &uiTilePoint);
	}
      else if( uiType == 1) // Outline Point
	{
	  // Partner Matrix
	  fscanf(file, "%s %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", 
		 pcFoo, &f0, &f1, &f2, &f3, &f4, &f5, &f6, &f7, &f8, &f9, &f10, 
		 &f11, &f12, &f13, &f14, &f15);
	  
	  mm[0]  = f0;
	  mm[1]  = f1;
	  mm[2]  = f2;
	  mm[3]  = f3;
	  mm[4]  = f4;
	  mm[5]  = f5;
	  mm[6]  = f6;
	  mm[7]  = f7;
	  mm[8]  = f8;
	  mm[9]  = f9;
	  mm[10] = f10;
	  mm[11] = f11;
	  mm[12] = f12;
	  mm[13] = f13;
	  mm[14] = f14;
	  mm[15] = f15;

	  m.Set(mm);
	}
      else
	{
	  ResetSymmetryGroup();
	  printf( "ES file \"%s\" has an invalid tilepoint type for sphere outline.\n", 
		  pcFile);
	  return;
	}
      pSG->CreateTilePoint(pt, uiType, uiTilePoint, 1, 0.0, m);
    }

  pSG->TesselateTileOutlinePoints();

  // Partners
  for( i = 0; i < uiSphereOutlinePoints; i++ )
    {
      // Partners
      fscanf(file, "%s %ld %ld", pcFoo, &uiID, &uiIDPartner);
      pSG->SetTilePointPartners(uiID, uiIDPartner);
      pSG->SetTileHeightPartnersFromOpenES(uiID, uiIDPartner);
    }
}

VOID ReadSphereInnerPoints( FILE *file, CHAR *pcFile )
{
  ASSERT( pSG != NULL );

  UINT_32 i, uiSphereInnerPoints, uiType, uiConnected;
  FLOAT fX, fY, fZ, fHeight;
  CHAR pcFoo[30];
  Point pt;
  Matrix m;

  pSG->ClearSphereInnerPoints();

  fscanf(file, "%s %ld", pcFoo, &uiSphereInnerPoints);

  for( i = 0; i < uiSphereInnerPoints; i++ )
    {
      // Point location and type
      fscanf(file, "%s %f %f %f %s %ld %s %ld %s %f", 
	     pcFoo, &fX, &fY, &fZ, pcFoo, &uiType, 
	     pcFoo, &uiConnected, pcFoo, &fHeight);

      pt.Set(fX, fY, fZ);
      ASSERT( uiType == 2 );

      pSG->CreateTilePoint(pt, uiType, 0, uiConnected, fHeight, m);
    }

  pSG->TesselateTileInnerPoints();
}

VOID ReadSphereHolePoints( FILE *file, CHAR *pcFile )
{
  ASSERT( pSG != NULL );

  UINT_32 i, uiSphereHolePoints, uiType;
  FLOAT fX, fY, fZ;
  CHAR pcFoo[30];
  Point pt;
  Matrix m;

  pSG->ClearSphereHolePoints();

  fscanf(file, "%s %ld", pcFoo, &uiSphereHolePoints);

  for( i = 0; i < uiSphereHolePoints; i++ )
    {
      // Point location and type
      fscanf(file, "%s %f %f %f %s %ld", 
	     pcFoo, &fX, &fY, &fZ, pcFoo, &uiType);

      pt.Set(fX, fY, fZ);
      ASSERT( uiType == 3 );

      pSG->CreateTilePoint(pt, uiType, 0, 0, 0, m);
    }
}

VOID ReadTileColors( FILE *file, CHAR *pcFile )
{
  ASSERT( pSG != NULL );

  UINT_32 i, uiTileColors, uiFoo;
  FLOAT r, g, b;
  CHAR pcFoo[30];
  Vector vColor;

  //WM
  CHAR filename[200];
  FILE *colorIndexFile, *paletteFile;
  UINT_32 idColor, uiIndexColors;

  //Reading the .pes file (paletteFile)
  fscanf(file, "%s %s", pcFoo, filename);
  paletteFile = fopen(filename, "ro");
  if(paletteFile == NULL) 
    printf ("Could not open paletteFile %s \n", filename);
  else 
    {
      fscanf(paletteFile, "%s %ld", pcFoo, &uiIndexColors);
      //printf("Index color = %ld \n", uiIndexColors);
      for (i = 0; i < uiIndexColors; i++) 
	{
	  fscanf(paletteFile, "%s %f %f %f",
		 pcFoo, &r, &g, &b);
	  //printf("%s %f %f %f \n", pcFoo, r, g, b);
	  vColor.Set(r, g, b);
	  pSG->SetColorIndex(i, vColor);
	  UpdateButtonColor(i, r, g, b);
	}
      fclose(paletteFile);
    }

  //Reading the .ces file (colorIndexFile)
  fscanf(file, "%s %s", pcFoo, filename);
  colorIndexFile = fopen(filename, "ro");
  if(colorIndexFile == NULL)
    printf ("Could not open colorIndexFile %s \n", filename);
  else 
    {
      fscanf(colorIndexFile, "%s %s %ld", pcFoo, pcFoo, &uiFoo);
      fscanf(colorIndexFile, "%s %ld", pcFoo, &uiTileColors);
      for (i = 0; i < uiTileColors; i++) 
	{
	  fscanf(colorIndexFile, "%s %ld", pcFoo, &idColor);
	  //printf("%s %ld \n", pcFoo, idColor);
	  pSG->SetIndexTile (i, idColor);
	}
      fclose(colorIndexFile);
    }
}

VOID OpenES( CHAR *pcFile )
{
  FILE *file;
  INT_32 iCode;

  ResetSymmetryGroup();

  // Reset Interface
  if( gbGUICreated )
    {
      ResetInterface();

      dtDrawType = dtSPHERE;
      sprintf(gpcBuffer, "set draw 0");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}
      
      ResetCamera();
    }

  // Read file
  file = fopen( pcFile, "ro" );
  {    
    // Symmetry Group and Tile
    ReadSymmetryGroup(file, pcFile); 

    // Tile
    if( pSG != NULL )
      {
	ReadSolid(file, pcFile);
      }

    // Sphere Outline Points and Partners
    if( pSG != NULL )
      {
	ReadSphereOutlinePoints(file, pcFile);
      }

    // Sphere Inner Points
    if( pSG != NULL )
      {
	ReadSphereInnerPoints(file, pcFile);
      }

    // Sphere Hole Points
    if( pSG != NULL )
      {
	ReadSphereHolePoints(file, pcFile);
      }

    // Tile Colors
    if( pSG != NULL )
      {
	ReadTileColors(file, pcFile);
      }
  }
  fclose(file);
}


//WM OpenPES
VOID OpenPES( CHAR *pcFile )
{
  FILE *file;
  CHAR pcFoo[30];
  FLOAT r, g, b;
  UINT_32 i, uiIndexColors;
  Vector vColor;

  // Read file
  file = fopen( pcFile, "ro" );
  {    
    // Tile Colors
    if( pSG != NULL )
      {
	fscanf(file, "%s %ld", pcFoo, &uiIndexColors);
	//printf("Index color = %ld \n", uiIndexColors);
	for (i = 0; i < uiIndexColors; i++) 
	  {
	    fscanf(file, "%s %f %f %f",
		   pcFoo, &r, &g, &b);
	    //printf("%s %f %f %f \n", pcFoo, r, g, b);
	    vColor.Set(r, g, b);
	    pSG->SetColorIndex(i, vColor);
	    UpdateButtonColor(i, r, g, b);
	    //WM: need to update the UI
	  }
      }
  }
  fclose(file);
}

//WM OpenCES
VOID OpenCES( CHAR *pcFile )
{
  FILE *file;
  UINT_32 i, uiTileColors, idColor, uiFoo;
  CHAR pcFoo[30];

  // Read file
  file = fopen( pcFile, "ro" );
  {    
    // Tile Colors
    if( pSG != NULL )
      {
	fscanf(file, "%s %s %ld", pcFoo, pcFoo, &uiFoo);
	fscanf(file, "%s %ld", pcFoo, &uiTileColors);
	//printf ("Tile color = %ld \n", uiTileColors);
	for (i = 0; i < uiTileColors; i++) 
	  {
	    fscanf(file, "%s %ld", pcFoo, &idColor);
	    //printf("%s %ld \n", pcFoo, idColor);
	    pSG->SetIndexTile (i, idColor);
	  }
      }
  }
  fclose(file);
  Togl_PostRedisplay( toglBall );
}

//WM HandleSetIndexColor
int CBHandleSetIndexColor( ClientData clientData, Tcl_Interp *pInterp, 
			   int argc, char **argv)
{
  UINT_32 colorIndex;

  ASSERT ( argc == 2 );

  colorIndex = atoi( argv[1] );
  //printf ("We are going to set index color: %ld\n", colorIndex);
  guiColorIndex = colorIndex;
  return TCL_OK;
}

//WM HandleUpdateIndexColor
int CBHandleUpdateIndexColor( ClientData clientData, Tcl_Interp *pInterp, 
			      int argc, char **argv)
{
  UINT_32 colorIndex;
  FLOAT r, g, b;
  Vector v;

  ASSERT ( argc == 5 );

  colorIndex = atoi( argv[1] );
  r = atoi( argv[2] ) / 65536.0;
  g = atoi( argv[3] ) / 65536.0;
  b = atoi( argv[4] ) / 65536.0;

  v.Set(r, g, b);
  //printf ("We are going to update index color: %ld\n", colorIndex);
  pSG->SetColorIndex(colorIndex, v);
  return TCL_OK;
}

int CBHandleOpenES( ClientData clientData, Tcl_Interp *pInterp, 
		    int argc, char **argv)
{
  ASSERT( argc == 1 );

  INT_32 iCode;
  UINT_32 uiLength;
  CHAR *pDirname;
  
  sprintf(gpcBuffer, "tk_getOpenFile -defaultextension .es -filetypes {{\"ES Files\" {.es}}}");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  uiLength = strlen(pInterp->result);
  if ( uiLength == 0 )
    {
      // The user hit Cancel
      return TCL_OK;
    }
  
  ASSERT( pInterp->result != NULL );
  gpcInputFile = new CHAR[uiLength + 1];
  ASSERT( gpcInputFile != NULL );
  strcpy(gpcInputFile, pInterp->result);
  
  // Change directory
  pDirname = GetFileDirPath ( gpcInputFile );
  sprintf(gpcBuffer, "cd %s", pDirname);
  delete[] pDirname;
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  OpenES( gpcInputFile );
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//WM handle openPES
int CBHandleOpenPES( ClientData clientData, Tcl_Interp *pInterp, 
		     int argc, char **argv)
{
  ASSERT( argc == 1 );

  INT_32 iCode;
  UINT_32 uiLength;
  CHAR *pDirname;
  
  sprintf(gpcBuffer, "tk_getOpenFile -defaultextension .pes -filetypes {{\"PES Files\" {.pes}}}");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  uiLength = strlen(pInterp->result);
  if ( uiLength == 0 )
    {
      // The user hit Cancel
      return TCL_OK;
    }
  
  ASSERT( pInterp->result != NULL );
  gpcInputFile = new CHAR[uiLength + 1];
  ASSERT( gpcInputFile != NULL );
  strcpy(gpcInputFile, pInterp->result);
  
  // Change directory
  pDirname = GetFileDirPath ( gpcInputFile );
  sprintf(gpcBuffer, "cd %s", pDirname);
  delete[] pDirname;
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  OpenPES( gpcInputFile );
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//WM handle openCES
// struct Togl *togl, int argc, char *argv[] 
int CBHandleOpenCES( ClientData clientData, Tcl_Interp *pInterp, 
		     int argc, char **argv)
{
  ASSERT( argc == 1 );

  INT_32 iCode;
  UINT_32 uiLength;
  CHAR *pDirname;
  
  sprintf(gpcBuffer, "tk_getOpenFile -defaultextension .ces -filetypes {{\"CES Files\" {.ces}}}");
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  uiLength = strlen(pInterp->result);
  if ( uiLength == 0 )
    {
      // The user hit Cancel
      return TCL_OK;
    }
  
  ASSERT( pInterp->result != NULL );
  gpcInputFile = new CHAR[uiLength + 1];
  ASSERT( gpcInputFile != NULL );
  strcpy(gpcInputFile, pInterp->result);
  
  // Change directory
  pDirname = GetFileDirPath ( gpcInputFile );
  sprintf(gpcBuffer, "cd %s", pDirname);
  delete[] pDirname;
  iCode = Tcl_Eval(pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );
  
  OpenCES( gpcInputFile );
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// Togl Callbacks

int CBHandleCamera( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 3 );

  INT_32 iAction;
  iAction = atoi( argv[2] );

  switch( iAction )
    {
    case 1: // Reset
      {
	ResetCamera();
      }
    break;
    default:
      {
	ASSERT( FALSE );      
      }
    break;
    }
  
  Togl_PostRedisplay( togl );

  return TCL_OK;
}

void animate()
{
  // I3D Video
  /*
  if( spinning ) 
    {
      add_quats(lastquat, curquat, curquat);

      float m[4][4];
      float mi[4][4];
      MATRIX4D mm;
      Matrix mBall;
      
      // rotation
      build_rotmatrix( m, lastquat );

      mm[0]  = m[0][0];
      mm[1]  = m[0][1];
      mm[2]  = m[0][2];
      mm[3]  = m[0][3];
      mm[4]  = m[1][0];
      mm[5]  = m[1][1];
      mm[6]  = m[1][2];
      mm[7]  = m[1][3];
      mm[8]  = m[2][0];
      mm[9]  = m[2][1];
      mm[10] = m[2][2];
      mm[11] = m[2][3];
      mm[12] = m[3][0];
      mm[13] = m[3][1];
      mm[14] = m[3][2];
      mm[15] = m[3][3];
      
      mBall.Set(mm);
      mBallRotate = mBall * mBallRotate;
      mBallCamera = mBallTranslate * mBallRotate * mBallScale;

      // inverse
      build_rotmatrix( mi, invlastquat );

      mm[0]  = mi[0][0];
      mm[1]  = mi[0][1];
      mm[2]  = mi[0][2];
      mm[3]  = mi[0][3];
      mm[4]  = mi[1][0];
      mm[5]  = mi[1][1];
      mm[6]  = mi[1][2];
      mm[7]  = mi[1][3];
      mm[8]  = mi[2][0];
      mm[9]  = mi[2][1];
      mm[10] = mi[2][2];
      mm[11] = mi[2][3];
      mm[12] = mi[3][0];
      mm[13] = mi[3][1];
      mm[14] = mi[3][2];
      mm[15] = mi[3][3];
      
      mBall.Set(mm);
      mBallInvRotate = mBallInvRotate * mBall;
    }
  
  Togl_PostRedisplay( toglBall );
  glFlush();

  if(!spinning) 
    {
      Tcl_CancelIdleCall(NULL, NULL);
    }
  else
    {
      Tcl_DoWhenIdle((Tcl_IdleProc *) animate, NULL);
    }
    */
}

VOID PickPoint(struct Togl *togl, int iWidgetX, int iWidgetY)
{
  if( (pSG == NULL) || 
      !((dtDrawType == dtSPHERE) || (dtDrawType == dtTRI_TILE) ||
	((dtDrawType == dtSOLID) && gbPaintMode) || 
	((dtDrawType == dtSOLID) && gbDropperMode)) )
    {
      return;
    }

  GLuint selectBuf[500];
  GLint hits;
  GLint viewport[4];
  unsigned int i, j, c;
  GLuint names;
  FLOAT fZMin, f1, f2;

  //////////////////////////////////////////////////
  // Draw to selection buffer

  glGetIntegerv(GL_VIEWPORT, viewport);
  glSelectBuffer(600, selectBuf);
  (void) glRenderMode(GL_SELECT);

  glInitNames();
  glPushName(SELECTED_NOTHING); // push on dummy name

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  {
    glLoadIdentity();

    gluPickMatrix((GLdouble)iWidgetX, (GLdouble)(viewport[3]-iWidgetY), 5.0, 5.0, viewport);
    glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
    glScalef( fBallScale, fBallScale, 1 );

    // Camera
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    {
      glLoadIdentity();	               
      glMultMatrixf(mBallCamera.Get());  // Apply the Crystal Ball Matrix

      if( dtDrawType == dtSPHERE )
	{
	  // This Render order is important!!!

	  // Render the sphere so point won't pick points on back of sphere
	  glCallList(GLuiSphere1_30);
	  
	  // Render and push names
	  glPushName(SELECTED_OUTLINE_EDGE);
	  {
	    pSG->RenderSelectionSphereTileOutlineEdges();
	  }
	  glPopName();
	  
	  glPushName(SELECTED_INNER_EDGE);
	  {
	    pSG->RenderSelectionSphereTileInnerEdges();
	  }
	  glPopName();
	  
	  glPushName(SELECTED_OUTLINE_POINT);
	  {
	    pSG->RenderSelectionSphereTileOutlinePoints();
	  }
	  glPopName();
	  
	  glPushName(SELECTED_INNER_POINT);
	  {
	    pSG->RenderSelectionSphereTileInnerPoints();
	  }
	  glPopName();
	  
	  glPushName(SELECTED_HOLE_POINT);
	  {
	    pSG->RenderSelectionSphereTileHolePoints();
	  }
	  glPopName();
	}
      else if( dtDrawType == dtTRI_TILE )
	{
	  glPushName(SELECTED_OUTLINE_POINT);
	  {
	    pSG->RenderSelectionSphereTileHeightOutlinePoints();
	  }
	  glPopName();

	  glPushName(SELECTED_INNER_POINT);
	  {
	    pSG->RenderSelectionSphereTileHeightInnerPoints();
	  }
	  glPopName();
	}
      else if( dtDrawType == dtSOLID )
	{
	  glPushName(SELECTED_TILE);
	  {
	    pSG->RenderSelectionSolid();
	  }
	  glPopName();
	}
    }
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
  }
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glFlush();

  hits = glRenderMode(GL_RENDER);

  // Draw to selection buffer
  //////////////////////////////////////////////////

  //////////////////////////////////////////////////
  // Process Hits

  //printf( "\nhits = %d\n", hits );

  guiSelectedType  = SELECTED_NOTHING;
  guiSelectedIndex = SELECTED_NOTHING;

  // Find which object was selected
  fZMin = INFINITY;
  for( i = 0, c = 0; i < (UINT_32)hits; i++ )
    {
      names = selectBuf[c];
      //printf(" names = %d: ", names);

      if( names > 10000 ) // this is some error, I'm not sure what
	{
	  printf("ERROR: selection names = %d: ", names);
	  return;
	}

      f1 = (float)selectBuf[c+1]/0x7fffffff;
      f2 = (float)selectBuf[c+2]/0x7fffffff;

      // Find nearest object (but take edges over sphere)
      if( (f1 < fZMin) ||
	  ((selectBuf[c+4] == SELECTED_OUTLINE_EDGE) && (f1 < fZMin + 0.01)) ||
	  ((selectBuf[c+4] == SELECTED_INNER_EDGE)   && (f1 < fZMin + 0.01)) )
	{
	  fZMin = f1;
	  	  
	  if( names == 3 )
	    {
	      guiSelectedType  = selectBuf[c+4];
	      guiSelectedIndex = selectBuf[c+5];
	    }
	}
      c += 3;

      for( j = 0; j < names; j++, c++ )
	{
	  //printf("%d ", selectBuf[c]);
	}

      //printf( "    zmin: %f  (%f)", f1, f2 );
      //printf( "\n" );
    }
  // Process Hits
  //////////////////////////////////////////////////

  //printf( "Selected: %lu %lu\n", guiSelectedType, guiSelectedIndex );
}

VOID MoveSelectedPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  ASSERT( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
	  (guiSelectedType == SELECTED_INNER_POINT) ||
	  (guiSelectedType == SELECTED_HOLE_POINT) );

  if( guiSelectedIndex == SELECTED_NOTHING )
    return;

  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      if( guiSelectedType == SELECTED_OUTLINE_POINT)
	{
	  pSG->MoveOutlinePoint(guiSelectedIndex, pt);
	}
      else if( guiSelectedType == SELECTED_INNER_POINT )
	{
	  pSG->MoveInnerPoint(guiSelectedIndex, pt);
	}
      else if( guiSelectedType == SELECTED_HOLE_POINT )
	{
	  pSG->MoveHolePoint(guiSelectedIndex, pt);
	}
      else
	{
	  ASSERT(FALSE);
	}
    }
}

VOID SnapInnerPointToSelected()
{
  ASSERT( (guiSelectedType == SELECTED_INNER_POINT) ||
	  (guiSelectedType == SELECTED_OUTLINE_POINT) );

  pSG->SnapInnerPointToSelected(guiCurrentSelectedInner, guiSelectedType, guiSelectedIndex);
}

VOID SlideSelectedPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  ASSERT( (guiSelectedType == SELECTED_INNER_POINT) || 
	  (guiSelectedType == SELECTED_OUTLINE_POINT) );

  FLOAT f;

  f = CalculateHeight( togl, iWidgetX, iWidgetY );
  pSG->AddHeightToPoints(f);
}

VOID InsertNewPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  if( guiSelectedIndex == SELECTED_NOTHING )
    return;

  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      if( guiSelectedType == SELECTED_OUTLINE_EDGE )
	{
	  pSG->InsertOutlinePoint(guiSelectedIndex, pt);
	  PickPoint( togl, iWidgetX, iWidgetY );
	}
      else if( guiSelectedType == SELECTED_INNER_EDGE )
	{
	  pSG->InsertInnerPoint(guiSelectedIndex, pt);
	  PickPoint( togl, iWidgetX, iWidgetY );
	}
    }
}

VOID AddNewInnerPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      gtptLastInnerPoint = pSG->AddInnerPoint(pt);
      PickPoint( togl, iWidgetX, iWidgetY );
    }
}

VOID AddNewInnerPointSnapToSelected()
{
  ASSERT( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
	  (guiSelectedType == SELECTED_INNER_POINT) );

  gtptLastInnerPoint = pSG->AddInnerPointCopy(guiSelectedType, guiSelectedIndex);
}

VOID AddNewHolePoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      pSG->AddHolePoint(pt);
      PickPoint( togl, iWidgetX, iWidgetY );
    }
}

VOID DeletePoint( struct Togl *togl )
{
  gbDeletingPoint = TRUE;

  if( guiSelectedType == SELECTED_OUTLINE_POINT )
    {
      pSG->DeleteOutlinePoint(guiSelectedIndex);
    }
  else if (guiSelectedType == SELECTED_INNER_POINT )
    {
      pSG->DeleteInnerPoint(guiSelectedIndex);
    }
  else if (guiSelectedType == SELECTED_HOLE_POINT )
    {
      pSG->DeleteHolePoint(guiSelectedIndex);
    }
  
  guiSelectedType  = SELECTED_NOTHING;
  guiSelectedIndex = SELECTED_NOTHING;
}

VOID SetHeightSelected(BOOL b)
{
  ASSERT( (guiSelectedType == SELECTED_INNER_POINT) ||
	  (guiSelectedType == SELECTED_OUTLINE_POINT) );

  pSG->SetHeightSelected( guiSelectedType, guiSelectedIndex, b );
}

VOID TogleHeightSelected()
{
  ASSERT( (guiSelectedType == SELECTED_INNER_POINT) ||
	  (guiSelectedType == SELECTED_OUTLINE_POINT) );

  pSG->TogleHeightSelected( guiSelectedType, guiSelectedIndex );
}

//WM FindColorIndex
VOID FindColorIndex() 
{
  UINT_32 indexColor;

  indexColor = pSG->GetColorIndex(guiSelectedIndex);
  if(indexColor == 99) 
    ExecuteTCLCommand("resetColorButton");
  else
    {
      sprintf(gpcBuffer2, "pushColorButton %ld", indexColor);
      ExecuteTCLCommand(gpcBuffer2);
    }
}

VOID PaintTile()
{
  ASSERT( guiSelectedType == SELECTED_TILE);

  // pass index also: guiColorIndex
  pSG->PaintTile(guiSelectedIndex, guiColorIndex);
}

int CBHandleKeyboard( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 6 );

  INT_32 iKey, iState, iWidgetX, iWidgetY;

  iKey     = (int) argv[2][0];
  iState   = atoi( argv[3] );
  iWidgetX = atoi( argv[4] );
  iWidgetY = atoi( argv[5] );

  switch( iKey )
    {
    case 'd':
      {
	//printf( "DeleteDown = %i\n", iState);
	//bDeleteKeyDown = iState;
      }
    break;
    default:
      {
      }
    break;    
    }

  return TCL_OK;
}

int CBHandleMouseButton( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 7 );

  SetStatusLabel( " " );

  INT_32 iButton, iPress, iWidgetX, iWidgetY, iTime, iWidth, iHeight;
  FLOAT fNormX, fNormY;
  Vector v;
 
  iButton  = atoi( argv[2] );
  iPress   = atoi( argv[3] );
  iWidgetX = atoi( argv[4] );
  iWidgetY = atoi( argv[5] );
  iTime    = atoi( argv[6] );

  iWidgetLastX = iWidgetX;
  iWidgetLastY = iWidgetY;

  iWidth = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;

  if( iButton == 1 ) // Left Button
    {
      gbLeftButtonDown = (iPress == 0);

      if( iPress == 0 ) // Button Press
	{
	  gbMouseMotion = FALSE;
	  spinning = 0;	 
	  
	  if( dtDrawType == dtSPHERE )
	    {
	      if( gbDeleteButtonDown )
		{
		  PickPoint(togl, iWidgetX, iWidgetY);
		  
		  if( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
		      (guiSelectedType == SELECTED_INNER_POINT) ||
		      (guiSelectedType == SELECTED_HOLE_POINT) )
		    {
		      DeletePoint( togl );
		      Togl_PostRedisplay( togl );
		    }
		}
	      else if( gbDrawInnerPoints )
		{
		  AddNewInnerPoint(togl, iWidgetX, iWidgetY);
		  Togl_PostRedisplay( togl );
		}
	      else if( gbDrawHolePoint )
		{
		  AddNewHolePoint(togl, iWidgetX, iWidgetY);
		  Togl_PostRedisplay( togl );
		}
	      else
		{
		  PickPoint(togl, iWidgetX, iWidgetY);
		  
		  if( ( guiSelectedType == SELECTED_OUTLINE_EDGE ) ||
		      ( guiSelectedType == SELECTED_INNER_EDGE ) )
		    {
		      InsertNewPoint(togl, iWidgetX, iWidgetY);
		      Togl_PostRedisplay( togl );
		    }
		}
	    }
	  else if( dtDrawType == dtTRI_TILE )
	    {
	      PickPoint(togl, iWidgetX, iWidgetY);

	      if( (guiSelectedType == SELECTED_INNER_POINT) ||
		  (guiSelectedType == SELECTED_OUTLINE_POINT) )
		{
		  if( !pSG->GetHeightSelected(guiSelectedType, guiSelectedIndex) )
		    {
		      SetHeightSelected(TRUE);
		      guiSelectedHeightIndex = guiSelectedIndex;
		      Togl_PostRedisplay( togl );
		    }
		}	      
	    }
	  else if( (dtDrawType == dtSOLID) && gbPaintMode )
	    {
	      PickPoint(togl, iWidgetX, iWidgetY);

	      if( guiSelectedType == SELECTED_TILE )
		{
		  PaintTile();
		  Togl_PostRedisplay( togl );
		}
	    }
	  else if( (dtDrawType == dtSOLID) && gbDropperMode )
	    {
	      //WM 
	      PickPoint(togl, iWidgetX, iWidgetY);
	      
	      if( guiSelectedType == SELECTED_TILE )
		{
		  FindColorIndex();
		}
	    }

	  iMouseX = iWidgetX;
	  iMouseY = iWidgetY;
	}
      else // Button Release
	{
	  if( spinning )
	    {
	      Tcl_DoWhenIdle((Tcl_IdleProc *) animate, NULL);
	    }

	  if( dtDrawType == dtSPHERE )
	    {
	      gbDeletingPoint = FALSE;
	      guiSelectedType  = SELECTED_NOTHING;
	      guiSelectedIndex = SELECTED_NOTHING;
	    }
	  else if( dtDrawType == dtTRI_TILE )
	    {
	      UINT_32 guiSelectedIndexPrev;

	      guiSelectedIndexPrev = guiSelectedIndex;

	      PickPoint(togl, iWidgetX, iWidgetY);
	      
	      if( (guiSelectedType == SELECTED_INNER_POINT) ||
		  (guiSelectedType == SELECTED_OUTLINE_POINT) )
		{
		  // click and release on same point
		  // this point wasn't just selected on the down click
		  // there was no motion
		  if( (guiSelectedIndexPrev == guiSelectedIndex) && 
		      (guiSelectedHeightIndex != guiSelectedIndex) &&
		      !gbMouseMotion )
		    {
		      TogleHeightSelected();
		      Togl_PostRedisplay( togl );
		    }
		}

	      guiSelectedHeightIndex = SELECTED_NOTHING;
	    }
	}
    }
  else if( iButton == 2 ) // Middle Button
    {
      gbMiddleButtonDown = (iPress == 0);

      if( iPress == 0 ) // Button Press
	{
	  if( dtDrawType == dtSPHERE )
	    {
	      if( gbLeftButtonDown )
		{
		  if( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
		      (guiSelectedType == SELECTED_INNER_POINT) ||
		      (guiSelectedType == SELECTED_HOLE_POINT) )
		    {
		      DeletePoint( togl );
		      Togl_PostRedisplay( togl );
		    }
		}
	      else
		{
		  PickPoint(togl, iWidgetX, iWidgetY);
		}
	      
	      /*
	      if( gbDrawInnerPoints )
		{
		  PickPoint(togl, iWidgetX, iWidgetY);
		  
		  if( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
		      (guiSelectedType == SELECTED_INNER_POINT) )
		    {
		      AddNewInnerPointSnapToSelected();
		      Togl_PostRedisplay( togl );
		    }
		  else
		    {
		      SetStatusLabel("Could not find an outline or inner point to snap to");
		    }
		}
		*/
	    }
	}
    }
  else if( iButton == 3 ) // Right Button
    {
      gbRightButtonDown = (iPress == 0);

      if (dtDrawType == dtSOLID) 
	{
	  Togl_PostRedisplay( togl );
	}

      /*
      if( iPress == 0 ) // Button Press
	{
	  iWidgetLastX = iWidgetX;
	  iWidgetLastY = iWidgetY;
	}
	*/
    }

  time = iTime;
  vBallNormMouseLast.Set( fNormX, fNormY, 0.0 );
      
  return TCL_OK;
}

int CBHandleMouseMotion( struct Togl *togl, int argc, char *argv[] )
{
  if( pSG == NULL )
    return TCL_OK;

  gbMouseMotion = TRUE;

  if( gbDeletingPoint )
    {
      return TCL_OK;
    }

  ASSERT( argc == 6 );

  INT_32 iButton, iWidgetX, iWidgetY, iTime, iWidth, iHeight;
  Vector vCurr, vMouseLast;
  FLOAT fNormX, fNormY;
  Matrix mBall;

  iButton  = atoi( argv[2] );
  iWidgetX = atoi( argv[3] );
  iWidgetY = atoi( argv[4] );
  iTime    = atoi( argv[5] );

  iWidth  = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;

  vCurr.Set( fNormX, fNormY, 0.0 );
  vMouseLast = vBallNormMouseLast;

  switch( iButton )
    {
    case 1:
      {
	// Left

	if( dtDrawType == dtSPHERE )
	  {
	    if( guiSelectedType != SELECTED_NOTHING )
	      {
		if( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
		    (guiSelectedType == SELECTED_INNER_POINT) ||
		    (guiSelectedType == SELECTED_HOLE_POINT) )
		  {
		    MoveSelectedPoint(togl, iWidgetX, iWidgetY);
		  }
		
		break; // Don't rotate
	      }
	  }
	else if( dtDrawType == dtTRI_TILE )
	  {
	    if( guiSelectedType != SELECTED_NOTHING )
	      {
		ASSERT( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
			(guiSelectedType == SELECTED_INNER_POINT) );

		SlideSelectedPoint(togl, iWidgetX, iWidgetY);
		break; // Don't rotate
	      }
	  }
	else if( dtDrawType == dtSOLID ) 
	  {
	    if( gbPaintMode )
	      {
		//break; // Don't rotate
	      }
	    if( gbDropperMode )
	      {
		//WM Is this the right thing to do??
	      }
	  }

	// Crystal Ball Rotate
	FLOAT m[4][4];
	FLOAT mi[4][4];
	FLOAT acceleration, fVelocity;
	MATRIX4D mm;
	int x, y;

	x = iWidgetX;
	y = iWidgetY;

	// speed = #pixels / millisecond
	fVelocity = abs(iMouseX - iWidgetX + iMouseY - iWidgetY) / (FLOAT)(iTime - time);
	acceleration = fVelocity - velocity / (FLOAT)(iTime - time);

	/*
	printf ("Delta Time: %d\n", iTime - time );
	printf ("Pixel's/Delta Time: %f\n",  
		abs(iMouseX - iWidgetX + iMouseY - iWidgetY) / (FLOAT)(iTime - time) );
		*/
	//printf ("Acceleration: %f\n",  acceleration );

	//if( fVelocity > 0.3)
	//if( acceleration > 0.3)
	if( acceleration > 0.1)
	  {
	    spinning = 1;
	  }
	else
	  {
	    spinning = 0;
	  }

	/*
	printf( "(%f %f) (%f %f)\n", (2.0*iMouseX - 500) / 500, (500 - 2.0*iMouseY) / 500,
		(2.0*x - 500) / 500,      (500 - 2.0*y) / 500);
		*/

	/*	
	trackball(lastquat, (2.0*(iMouseX-iCenterX)-500)/500, (500-2.0*(iMouseY-iCenterY))/500,
		                 (2.0*(x-iCenterX)-500)/500,    (500 - 2.0*(y-iCenterY))/500);

	trackball(invlastquat, (2.0*(x-iCenterX)-500)/500,    (500 - 2.0*(y-iCenterY))/500,
		          (2.0*(iMouseX-iCenterX)-500)/500, (500-2.0*(iMouseY-iCenterY))/500);
			  */

	trackball(fBallScale,
		  lastquat, 
		  (2.0*(iMouseX-iCenterX)-iToglWidth)/iToglWidth, 
		  (iToglHeight-2.0*(iMouseY-iCenterY))/iToglHeight,
		  (2.0*(x-iCenterX)-iToglWidth)/iToglWidth,    
		  (iToglHeight - 2.0*(y-iCenterY))/iToglHeight);

	trackball(fBallScale,
		  invlastquat, 
		  (2.0*(x-iCenterX)-iToglWidth)/iToglWidth,    
		  (iToglHeight - 2.0*(y-iCenterY))/iToglHeight,
		  (2.0*(iMouseX-iCenterX)-iToglWidth)/iToglWidth, 
		  (iToglHeight-2.0*(iMouseY-iCenterY))/iToglHeight);
		                 
	build_rotmatrix( m, lastquat );
	build_rotmatrix( mi, invlastquat );

	// rotation
	mm[0] = m[0][0];
	mm[1] = m[0][1];
	mm[2] = m[0][2];
	mm[3] = m[0][3];
	mm[4] = m[1][0];
	mm[5] = m[1][1];
	mm[6] = m[1][2];
	mm[7] = m[1][3];
	mm[8] =  m[2][0];
	mm[9] =  m[2][1];
	mm[10] = m[2][2];
	mm[11] = m[2][3];
	mm[12] = m[3][0];
	mm[13] = m[3][1];
	mm[14] = m[3][2];
	mm[15] = m[3][3];

	mBall.Set(mm);
	mBallRotate = mBall * mBallRotate;

	// inverse
	mm[0]  = mi[0][0];
	mm[1]  = mi[0][1];
	mm[2]  = mi[0][2];
	mm[3]  = mi[0][3];
	mm[4]  = mi[1][0];
	mm[5]  = mi[1][1];
	mm[6]  = mi[1][2];
	mm[7]  = mi[1][3];
	mm[8]  = mi[2][0];
	mm[9]  = mi[2][1];
	mm[10] = mi[2][2];
	mm[11] = mi[2][3];
	mm[12] = mi[3][0];
	mm[13] = mi[3][1];
	mm[14] = mi[3][2];
	mm[15] = mi[3][3];

	mBall.Set(mm);
	mBallInvRotate = mBallInvRotate * mBall;

	iMouseX = iWidgetX;
	iMouseY = iWidgetY;
	velocity = fVelocity;
      }
    break;
    case 2:
      {
	// Middle

	if( dtDrawType == dtSPHERE )
	  {
	    if( guiSelectedType != SELECTED_NOTHING )
	      {
		if( (guiSelectedType == SELECTED_INNER_POINT) )
		  {
		    guiCurrentSelectedInner = guiSelectedIndex;

		    // Snap check
		    PickPoint(togl, iWidgetX, iWidgetY);
		  
		    if( (guiSelectedType == SELECTED_OUTLINE_POINT) ||
			(guiSelectedType == SELECTED_INNER_POINT) )
		      {
			SnapInnerPointToSelected();

			guiSelectedType = SELECTED_INNER_POINT;
			guiSelectedIndex = guiCurrentSelectedInner;

			Togl_PostRedisplay( togl );
		      }
		    else // don't snap just move
		      {
			guiSelectedType = SELECTED_INNER_POINT;
			guiSelectedIndex = guiCurrentSelectedInner;

			MoveSelectedPoint(togl, iWidgetX, iWidgetY);
		      }

		    guiCurrentSelectedInner = SELECTED_NOTHING;
		  }		
		break; // Don't scale
	      }
	  }	

	/*
	if( gbLeftButtonDown && (guiSelectedType == SELECTED_OUTLINE_POINT) )
	  {
	    break;
	  }
	  */

	
	/*
	FLOAT fDolly;

	fDolly = 2.0*((FLOAT)fNormY-vBallNormMouseLast[Y]);
	v.Set( 0.0, 0.0, fDolly);
	mBall.SetTranslate(v);
	*/

	FLOAT fScale, fNormCenterX, fNormCenterY;
	Point pt;
	Matrix m;

	fScale = pow(2, 5.0*((FLOAT)fNormY-vMouseLast[Y]));

	// Scaling the World
	//mBall.SetScale(fScale);
	//mBallScale = mBall * mBallScale;

	// Scaling the projection matrix (so light ok)
	fBallScale *= fScale;

	//gbZoomIn = (fBallScale > 3.0);
	//printf( "fBallScale = %f\n", fBallScale);

	// Model center -> norm space -> widget space
	pt.Set( mBallTranslate.Get(0, 3), mBallTranslate.Get(1, 3), 0 );
	fNormCenterX = (2.0/5.0)*fBallScale * pt[X];
	fNormCenterY = (2.0/5.0)*fBallScale * pt[Y];	
	iCenterX = (int)( iToglWidth * fNormCenterX);
	iCenterY = (int)(-iToglHeight * fNormCenterY);
      }
    break;
    case 3:
      {
	// Right


	// Translate
	FLOAT fX, fY;
	Vector v;

	fX = Norm2ModelSpace( (FLOAT)fNormX-vBallNormMouseLast[X] );
	fY = Norm2ModelSpace( (FLOAT)fNormY-vBallNormMouseLast[Y] );

	v.Set( fX, fY, 0.0 );
	mBall.SetTranslate(v);

	mBallTranslate = mBall * mBallTranslate;
	iCenterX += (iWidgetX-iWidgetLastX);
	iCenterY += (iWidgetY-iWidgetLastY);
      }
    break;
    default:
      {
	ASSERT( FALSE );      
      }
    break;
    }
  
  time = iTime;
  //mBallCamera = mBall * mBallCamera;
  mBallCamera = mBallTranslate * mBallRotate * mBallScale;
  vBallNormMouseLast.Set( fNormX, fNormY, 0.0 );
  iWidgetLastX = iWidgetX;
  iWidgetLastY = iWidgetY;


  Togl_PostRedisplay( togl );

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////////


// This is called by Tcl/Tk when the widget has been realized.  
// Here's where one may do some one-time context setup or initializations.
void CBHandleToglCreate( struct Togl *togl )
{
  toglBall = togl;
  
  glShadeModel(GL_FLAT);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_COLOR_MATERIAL);

  glCullFace(GL_BACK);

  trackball(fBallScale, curquat, 0.0, 0.0, 0.0, 0.0);
}

// This is called by Tcl/Tk when the widget has been resized.  
// Typically, we call glViewport and perhaps setup the projection matrix.
void CBHandleToglReshape( struct Togl *togl )
{
  int width = Togl_Width( togl );
  int height = Togl_Height( togl );
  
  glViewport( 0, 0, width, height );
  
  // Set up projection transform
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
  
  // Change back to model view transform for rendering
  glMatrixMode(GL_MODELVIEW);
}

// This is called by Tcl/Tk when the widget's contents have to be redrawn.  
// Typically, we clear the color and depth buffers, render our objects, then 
// swap the front/back color buffers.
void CBHandleToglDisplay( struct Togl *togl )
{
  //glClearColor(1.0, 1.0, 1.0, 0.0); // Background Color
  glClearColor(0.0, 0.0, 0.0, 0.0); // Background Color
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Lights
  FLOAT light_ambient[]  = { 0.2, 0.2, 0.2, 1.0 };
  FLOAT light_diffuse[]  = { 1.0, 1.0, 1.0, 1.0 };
  FLOAT light_specular[] = { 0.5, 0.5, 0.5, 1.0 };
  FLOAT position[] = { 5.0, 5.0, 5.0, 0.0 };   // Directional

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, position);

  // Camera & Scale
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
  glScalef( fBallScale, fBallScale, 1 );
 
  // Model & Crystal Ball
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();	                 

  glEnable( GL_DEPTH_TEST );
  {
    /*
    // Draw Circle Center
    glPushMatrix();
    {
      FLOAT fX, fY;      
      fX = (float)iCenterX / 500;
      fY = (float)iCenterY / 500;
      fX = Norm2ModelSpace( fX );
      fY = Norm2ModelSpace( fY );
      glColor3f(1, 0, 0);
      glTranslatef( fX, -fY, 1.05);
      glCallList(GLuiSphere02_8);
    }
    glPopMatrix();
    */

    glMultMatrixf(mBallCamera.Get());  // Apply the Crystal Ball Matrix

    if( pSG != NULL )
      {
	switch( dtDrawType )
	  {
	  case dtSPHERE:
	    {
	      glEnable(GL_CULL_FACE);
	      {
		pSG->RenderBall();	      
	      }
	      glDisable(GL_CULL_FACE);

	      /*
		glCallList(GLuiSphere1_30);
	      pSG->RenderSelectionSphereTileOutlinePoints();
	      pSG->RenderSelectionSphereTileOutlineEdges();
	      pSG->RenderSelectionSphereTileInnerPoints();
	      pSG->RenderSelectionSphereTileInnerEdges();
	      */
	    }
	  break;
	  case dtPOLYHEDRA:
	    {
	      glEnable(GL_CULL_FACE);
	      {
		pSG->RenderPolyhedra();
	      }
	      glDisable(GL_CULL_FACE);
	    }
	  break;
	  case dtSTEREO:
	    {
	      pSG->RenderStereographicPoints();
	    }
	  break;
	  case dtTRI_TILE:
	    {
	      pSG->RenderSphereTriangulation();
	    }
	  break;
	  case dtSOLID:
	    {
	      glEnable(GL_CULL_FACE);
	      {
		pSG->RenderSolid();
	      }
	      glDisable(GL_CULL_FACE);
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }
	  break;
	  }
      }
  }
  glDisable( GL_DEPTH_TEST );

  glMatrixMode(GL_MODELVIEW);  
  glLoadIdentity();  
  Togl_SwapBuffers( togl );
}

// Called by Tk_Main() to let me initialize the modules (Togl) I will need.
int Tcl_AppInit( Tcl_Interp *pInterp )
{
  int iCode;

  pTclInterp = pInterp;

  // Initialize Tcl, Tk, and the Togl widget module.
  if (Tcl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Tk_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Togl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

#ifdef WIN32
  // Set up a console window. Delete the following statement if you do not need that.
  if (TkConsoleInit(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
#endif // WIN32
  
  // Specify the C callback functions for widget creation, display, and reshape.
  Togl_CreateFunc(  CBHandleToglCreate  );
  Togl_DisplayFunc( CBHandleToglDisplay );
  Togl_ReshapeFunc( CBHandleToglReshape );

  // Make a new Togl widget command so the Tcl code can set a C variable.
  Togl_CreateCommand( "keyboard", CBHandleKeyboard );
  Togl_CreateCommand( "mouseButton", CBHandleMouseButton );
  Togl_CreateCommand( "mouseMotion", CBHandleMouseMotion );
  Togl_CreateCommand( "camera", CBHandleCamera );


  // Create Tcl Commands
  if( Tcl_CreateCommand(pInterp, "guiCreated", CBHandleGUICreated, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setSymmetryGroup", CBHandleSetSymmetryGroup, NULL, NULL) 
      == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setDraw", CBHandleSetDraw, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setRadius", CBHandleSetRadius, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setThickness", CBHandleSetThickness, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setExtrude", CBHandleSetExtrude, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setShell", CBHandleSetShell, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setSolidTiles", CBHandleSetSolidTiles, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setPaintMode", CBHandleSetPaintMode, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  //WM dropperMode
  if( Tcl_CreateCommand(pInterp, "setDropperMode", CBHandleSetDropperMode, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setPaintColor", CBHandleSetPaintColor, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "drawGrid", CBHandleDrawGrid, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "drawNP", CBHandleDrawNorthPole, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "innerPoints", CBHandleInnerPoints, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "holePoint", CBHandleHolePoint, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "deleteDown", CBHandleDeleteDown, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "deselectInner", CBHandleDeselectInner, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "triangulate", CBHandleTriangulate, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "zoom", CBHandleZoom, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputSTL", CBHandleOutputSTL, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputVRML", CBHandleOutputVRML, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputES", CBHandleOutputES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  //WM
  if( Tcl_CreateCommand(pInterp, "openPES", CBHandleOpenPES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputPES", CBHandleOutputPES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "openCES", CBHandleOpenCES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }

  if( Tcl_CreateCommand(pInterp, "outputCES", CBHandleOutputCES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setIndexColor", CBHandleSetIndexColor, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "updateIndexColor", CBHandleUpdateIndexColor, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  //WM Finish adding new tcl_command
  if( Tcl_CreateCommand(pInterp, "openES", CBHandleOpenES, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "incrUI", CBHandleIncrUI, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "decrUI", CBHandleDecrUI, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "tri", CBHandleTriFiles, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }

  // Specify a user-specific startup file to invoke if the application is run interactively.

#ifdef WIN32
  iCode = Tcl_EvalFile(pInterp, "main_nt.tcl"); 
#else
  iCode = Tcl_EvalFile(pInterp, "main.tcl"); 
#endif

  if ( iCode != TCL_OK )
    {
#ifdef WIN32
      sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error opening Main configure file: \\\"main_nt.tcl\\\"\\n Error Message: %s\n Exiting program.\"", 
	      pInterp->result);
#else
      sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error opening Main configure file: \\\"main.tcl\\\"\\n Error Message: %s\n Exiting program.\"", 
	      pInterp->result);
#endif
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      return TCL_ERROR;
    }

  Init();
  
  return TCL_OK;
}

int main( int argc, char *argv[] )
{
  Tk_Main( argc, argv, Tcl_AppInit );
  return 0;
}

