#include "ooctahedron.h"

//////////////////////////////////////////////////////////////////////
// OOctahedron Class
//

OOctahedron::OOctahedron()
{
  m_fDihedralAngle = acos(-1.0/3.0); // 109.47
}

OOctahedron::~OOctahedron()
{
}

VOID OOctahedron::InitFaceGrid()
{
  Vector v1, v2, v3;
  FLOAT y1;

  y1 = sqrt(2);

  v1.Set( 0.0,  y1, 0.0);
  v2.Set(-1.0, 0.0, 1.0);
  v3.Set( 1.0, 0.0, 1.0);
  
  v1.Normalize();
  v2.Normalize();
  v3.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
}

VOID OOctahedron::InitSphereFaceGrid()
{
  FLOAT y1;
  Point pt, pt1, pt2, pt3;
  Vector v;

  // 3D
  y1 = sqrt(2);

  pt1.Set( 0.0,  y1, 0.0);
  pt2.Set(-1.0, 0.0, 1.0);
  pt3.Set( 1.0, 0.0, 1.0);

  // [pt1 -> pt2 -> pt3]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt1);
}

VOID OOctahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();

  if( (GetType() == estOOTile1) ||
      (GetType() == estOOTile2) )
    {
      tpt2->SetHeightPartner(tpt4);
      tpt4->SetHeightPartner(tpt2);
    }
  else if( GetType() == estOOTile3 )
    {
      tpt1->SetHeightPartner(tpt3);
      tpt3->SetHeightPartner(tpt1);
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID OOctahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;
  
  m_listFaceTransformMatrix.RemoveAll();
  
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();

    // Face 1
    glPushMatrix();
    {
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glRotatef(90, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glRotatef(180, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
    // Face 4
    glPushMatrix();
    {
      glRotatef(270, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();


    glScalef(-1, -1, 1);
    
    // Face 5
    glPushMatrix();
    {
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
    // Face 6
    glPushMatrix();
    {
      glRotatef(90, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 7
    glPushMatrix();
    {
      glRotatef(180, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
    // Face 8
    glPushMatrix();
    {
      glRotatef(270, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();    
  }
  glPopMatrix();
}

VOID OOctahedron::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;
  Vector v;

  v.Set(0, sqrt(2)/(2*sqrt(3)), 1/sqrt(3));
  v.Normalize();

  m_listTileTransformMatrix.RemoveAll();

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(120));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(-120));
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID OOctahedron::FindMatrices(Matrix &/*m1*/, Matrix &/*m2*/, TilePoint */*tpt1*/)
{
  ASSERT(FALSE);
}

///////////////////////////////////////////////////////////////////////////////

VOID OOctahedron::OutputES(FILE *fpOutputFile, CHAR *paletteFilename,
			   CHAR *colorFilename)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "octahedron\n\n" );

  m_pTile->OutputES(fpOutputFile);

  fprintf(fpOutputFile, "palettefile %s\n", paletteFilename);
  fprintf(fpOutputFile, "colorindexfile %s\n", colorFilename);
  //OutputESColor(fpOutputFile);
}
