#include "otetrahedron.h"


//////////////////////////////////////////////////////////////////////
// OTetrahedron Class
//

OTetrahedron::OTetrahedron()
{
  m_fDihedralAngle = acos(1.0/3.0); // 70.52
}

OTetrahedron::~OTetrahedron()
{
}

VOID OTetrahedron::InitFaceGrid()
{
  FLOAT y1, y2, z;
  Vector v1, v2, v3;

  // 2D
  y1 = 1/sqrt(3);
  y2 = -1/(2.0*sqrt(3));
  z  = sqrt(6)/12.0;

  v1.Set( 0.0, y1, z);
  v2.Set(-0.5, y2, z);
  v3.Set( 0.5, y2, z);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
}

VOID OTetrahedron::InitSphereFaceGrid()
{
  FLOAT y1, y2, z;
  Point pt, pt1, pt2, pt3;
  Vector v;

  // 3D
  y1 = -1/(2.0*sqrt(3));
  y2 = 1/sqrt(3);
  z  = sqrt(6)/12.0;

  pt1.Set( 0.0, y2, z);
  pt2.Set(-0.5, y1, z);
  pt3.Set( 0.5, y1, z);

  // [pt1 -> pt2 -> pt3]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt1);
}

VOID OTetrahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();

  if( (GetType() == estOTTile1) ||
      (GetType() == estOTTile2) )
    {
      tpt2->SetHeightPartner(tpt4);
      tpt4->SetHeightPartner(tpt2);
    }
}

VOID OTetrahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;

  m_listFaceTransformMatrix.RemoveAll();

  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();
    
    // Face 1
    glPushMatrix();
    {
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glRotatef( 120, 0, 0, 1);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 4
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glRotatef( -120, 0, 0, 1);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
  }
  glPopMatrix();
}

VOID OTetrahedron::InitTileMatrices()
{
  Matrix *mm;

  m_listTileTransformMatrix.RemoveAll();

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(120.0));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(-120.0));
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID OTetrahedron::FindMatrices(Matrix &/*m1*/, Matrix &/*m2*/, TilePoint */*tpt1*/)
{
  ASSERT(FALSE);
}

///////////////////////////////////////////////////////////////////////////////

VOID OTetrahedron::OutputES(FILE *fpOutputFile, 
			    CHAR *paletteFilename,
			    CHAR *colorFilename)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "otetrahedron\n\n" );

  m_pTile->OutputES(fpOutputFile);

  fprintf(fpOutputFile, "palettefile %s\n", paletteFilename);
  fprintf(fpOutputFile, "colorindexfile %s\n", colorFilename);
  //OutputESColor(fpOutputFile);
}
