#include "pentadodecahedron.h"

//////////////////////////////////////////////////////////////////////
// PentaDodecahedron Class
//

PentaDodecahedron::PentaDodecahedron()
{
  m_fDihedralAngle = acos(-sqrt(5.0)/5.0); // 116.57

  Vector v1, v2, v3, v4, v5, v6;
  FLOAT f;
  
  v1.Set(            0.0,   GOLDENRATIO, INVGOLDENRATIO);
  v2.Set(           -1.0,           1.0,            1.0);
  v3.Set(-INVGOLDENRATIO,           0.0,    GOLDENRATIO);
  v4.Set( INVGOLDENRATIO,           0.0,    GOLDENRATIO);
  v5.Set( (INVGOLDENRATIO+1.0)/2.0, 0.5, (GOLDENRATIO+1.0)/2.0);
  v6.Set(            1.0,           1.0,            1.0);

  f = 1.0 / v1.Magnitude();

  v1 *= f;
  v2 *= f;
  v3 *= f;
  v4 *= f;
  v5 *= f;
  v6 *= f;

  // 2D
  m_pt2DTilePoint1.Set(v1[X], v1[Y], v1[Z]);
  m_pt2DTilePoint2.Set(v2[X], v2[Y], v2[Z]);
  m_pt2DTilePoint3.Set(v3[X], v3[Y], v3[Z]);
  m_pt2DTilePoint4.Set(v4[X], v4[Y], v4[Z]);
  m_pt2DTilePoint5.Set(v5[X], v5[Y], v5[Z]);
  m_pt2DTilePoint6.Set(v6[X], v6[Y], v6[Z]);

  Init();
}

PentaDodecahedron::~PentaDodecahedron()
{
}

VOID PentaDodecahedron::InitTileGrid()
{
  // 2D 
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint1 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint2 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint3 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint4 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint5 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint6 );
}

VOID PentaDodecahedron::InitFaceGrid()
{
  Vector v1, v2, v3, v4, v5;

  v1.Set(            0.0, GOLDENRATIO, INVGOLDENRATIO);
  v2.Set(           -1.0,         1.0,            1.0);
  v3.Set(-INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  v4.Set( INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  v5.Set(            1.0,         1.0,            1.0);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();
  v4.Normalize();
  v5.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
  m_pTile->InsertFaceGridPoint( v4[X], v4[Y], v4[Z]);
  m_pTile->InsertFaceGridPoint( v5[X], v5[Y], v5[Z]);
}

VOID PentaDodecahedron::InitSphereTileGrid()
{
  // [pt1 -> pt2 -> pt3 -> pt4 -> pt5 -> pt6]
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint1, m_pt2DTilePoint2);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint2, m_pt2DTilePoint3);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint3, m_pt2DTilePoint4);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint4, m_pt2DTilePoint5);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint5, m_pt2DTilePoint6);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint6, m_pt2DTilePoint1);
}

VOID PentaDodecahedron::InitSphereFaceGrid()
{
  Point pt1, pt2, pt3, pt4, pt5;

  // 3D
  pt1.Set(            0.0, GOLDENRATIO, INVGOLDENRATIO);
  pt2.Set(           -1.0,         1.0,            1.0);
  pt3.Set(-INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  pt4.Set( INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  pt5.Set(            1.0,         1.0,            1.0);

  // [pt1 -> pt2 -> pt3 -> pt4 -> pt5]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt4);
  AddSphereFaceGridTesselatedPoints(pt4, pt5);
  AddSphereFaceGridTesselatedPoints(pt5, pt1);
}

VOID PentaDodecahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4, *tpt5, *tpt6;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );
  tpt5 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint5, tpTILE, 5 );
  tpt6 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint6, tpTILE, 6 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt6);
  tpt5->SetPartner(tpt5);
  tpt6->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();
}

VOID PentaDodecahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;
  
  m_listFaceTransformMatrix.RemoveAll();
  
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();

    /////////////////////////////////////////////////////////////////////////
    // Top

    // Face 1
    glPushMatrix();
    {
      // Identity
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glRotatef(120, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glRotatef(240, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 4
    glPushMatrix();
    {
      glRotatef(180, (INVGOLDENRATIO+1.0)/2.0, 0.5, (GOLDENRATIO+1.0)/2.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 5
    glPushMatrix();
    {
      glRotatef(120, -INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 6
    glPushMatrix();
    {
      glRotatef(240, -INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////

    // Face 7
    glPushMatrix();
    {
      glRotatef(120, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glRotatef(72, 0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO+2*GOLDENRATIO + 2.0)/5.0 );
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 8
    glPushMatrix();
    {
      glRotatef(240, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glRotatef(72, 0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO+2*GOLDENRATIO + 2.0)/5.0 );
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 9
    glPushMatrix();
    {
      glRotatef(240, -INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glRotatef(120, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////

    // Face 10
    glPushMatrix();
    {
      glRotatef(120, INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glRotatef(240, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glRotatef(72, 0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO+2*GOLDENRATIO + 2.0)/5.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 11
    glPushMatrix();
    {
      glRotatef(240, 1.0, 1.0, 1.0);
      glRotatef(120, -INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glRotatef(-72, 0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO+2*GOLDENRATIO + 2.0)/5.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 12
    glPushMatrix();
    {
      glRotatef(180, 1.0, 0.0, 0.0);
      glRotatef(3*72, 0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO+2*GOLDENRATIO + 2.0)/5.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
  }
  glPopMatrix();
}

VOID PentaDodecahedron::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;
  Vector v;

  m_listTileTransformMatrix.RemoveAll();

  v.Set(0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO + 2*GOLDENRATIO + 2.0)/5.0 );

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID PentaDodecahedron::FindMatrices(Matrix &m1, Matrix &m2, TilePoint *tpt1)
{
  Vector v;

  if( tpt1->GetType() != tpTILE )
    {
      m1 = tpt1->GetPartnerMatrix();
      m2 = (tpt1->GetPartner())->GetPartnerMatrix();
    }
  else if( tpt1->GetTilePoint() == 1 )
    {
      v.Set(0.0,   GOLDENRATIO, INVGOLDENRATIO);
      m1.SetRotate(v, DEGTORAD(120));
      m2.SetRotate(v, DEGTORAD(-120));
    }
  else if( tpt1->GetTilePoint() == 2 )
    {
      v.Set(-INVGOLDENRATIO, 0.0, GOLDENRATIO);
      m1.SetRotate(v, DEGTORAD(-120));
      m2.SetRotate(v, DEGTORAD(120));
    }
  else if( tpt1->GetTilePoint() == 3 )
    {
      v.Set(-INVGOLDENRATIO, 0.0, GOLDENRATIO);
      m1.SetRotate(v, DEGTORAD(120));
      m2.SetRotate(v, DEGTORAD(-120));
    }
  else if( tpt1->GetTilePoint() == 4 )
    {
      v.Set((INVGOLDENRATIO+1.0)/2.0, 0.5, (GOLDENRATIO+1.0)/2.0);
      m1.SetRotate(v, DEGTORAD(-180));
      m2.SetRotate(v, DEGTORAD(180));
    }
  else if( tpt1->GetTilePoint() == 5 )
    {
      v.Set((INVGOLDENRATIO+1.0)/2.0, 0.5, (GOLDENRATIO+1.0)/2.0);
      m1.SetRotate(v, DEGTORAD(180));
      m2.SetRotate(v, DEGTORAD(-180));
    }
  else if( tpt1->GetTilePoint() == 6 )
    {
      v.Set(0.0,   GOLDENRATIO, INVGOLDENRATIO);
      m1.SetRotate(v, DEGTORAD(-120));
      m2.SetRotate(v, DEGTORAD(120));
    }
  else
    {
      ASSERT(FALSE);
    }
}

///////////////////////////////////////////////////////////////////////////////

VOID PentaDodecahedron::OutputES(FILE *fpOutputFile, CHAR *paletteFilename,
				 CHAR *colorFilename)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "pentadodecahedron tile 0\n\n" );

  m_pTile->OutputES(fpOutputFile);

  fprintf(fpOutputFile, "palettefile %s\n", paletteFilename);
  fprintf(fpOutputFile, "colorindexfile %s\n", colorFilename);
  //OutputESColor(fpOutputFile);
}
