#include "ocube.h"

//////////////////////////////////////////////////////////////////////
// OCube Class
//

OCube::OCube()
{
  m_fDihedralAngle = DEGTORAD(90); // 90
}

OCube::~OCube()
{
}

VOID OCube::InitFaceGrid()
{
  Vector v1, v2, v3, v4;

  v1.Set( 1.0,  1.0, 1.0);
  v2.Set(-1.0,  1.0, 1.0);
  v3.Set(-1.0, -1.0, 1.0);
  v4.Set( 1.0, -1.0, 1.0);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();
  v4.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
  m_pTile->InsertFaceGridPoint( v4[X], v4[Y], v4[Z]);
}

VOID OCube::InitSphereFaceGrid()
{
  Point pt, pt1, pt2, pt3, pt4;

  // 3D
  pt1.Set( 0.5,  0.5, 0.5);
  pt2.Set(-0.5,  0.5, 0.5);
  pt3.Set(-0.5, -0.5, 0.5);
  pt4.Set( 0.5, -0.5, 0.5);

  // [pt1 -> pt2 -> pt3 -> pt4]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt4);
  AddSphereFaceGridTesselatedPoints(pt4, pt1);
}

VOID OCube::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();

  if( (GetType() == estOCTile1) ||
      (GetType() == estOCTile2) )
    {
      tpt2->SetHeightPartner(tpt4);
      tpt4->SetHeightPartner(tpt2);
    }
  else if( GetType() == estOCTile3 )
    {
      tpt1->SetHeightPartner(tpt3);
      tpt3->SetHeightPartner(tpt1);
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID OCube::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;
  
  m_listFaceTransformMatrix.RemoveAll();
  
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();

    // Face 1
    glPushMatrix();
    {
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glRotatef(90, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glRotatef(180, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
    // Face 4
    glPushMatrix();
    {
      glRotatef(270, 0, 1, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 5
    glPushMatrix();
    {
      glRotatef(90, 1, 0, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
    // Face 6
    glPushMatrix();
    {
      glRotatef(-90, 1, 0, 0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
  }
  glPopMatrix();
}

VOID OCube::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;

  m_listTileTransformMatrix.RemoveAll();

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(90));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(180));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 4
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(270));
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID OCube::FindMatrices(Matrix &/*m1*/, Matrix &/*m2*/, TilePoint */*tpt1*/)
{
  ASSERT(FALSE);
}

///////////////////////////////////////////////////////////////////////////////

VOID OCube::OutputES(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "ocube\n\n" );

  m_pTile->OutputES(fpOutputFile);

  OutputESColor(fpOutputFile);
}
