#include "ododecahedron.h"

//////////////////////////////////////////////////////////////////////
// ODodecahedron Class
//

ODodecahedron::ODodecahedron()
{
  m_fDihedralAngle = acos(-sqrt(5.0)/5.0); // 116.57
}

ODodecahedron::~ODodecahedron()
{
}

VOID ODodecahedron::InitFaceGrid()
{
  Vector v1, v2, v3, v4, v5;

  v1.Set(            0.0, GOLDENRATIO, INVGOLDENRATIO);
  v2.Set(           -1.0,         1.0,            1.0);
  v3.Set(-INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  v4.Set( INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  v5.Set(            1.0,         1.0,            1.0);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();
  v4.Normalize();
  v5.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
  m_pTile->InsertFaceGridPoint( v4[X], v4[Y], v4[Z]);
  m_pTile->InsertFaceGridPoint( v5[X], v5[Y], v5[Z]);
}

VOID ODodecahedron::InitSphereFaceGrid()
{
  Point pt1, pt2, pt3, pt4, pt5;

  // 3D
  pt1.Set(            0.0, GOLDENRATIO, INVGOLDENRATIO);
  pt2.Set(           -1.0,         1.0,            1.0);
  pt3.Set(-INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  pt4.Set( INVGOLDENRATIO,         0.0,    GOLDENRATIO);
  pt5.Set(            1.0,         1.0,            1.0);

  // [pt1 -> pt2 -> pt3 -> pt4 -> pt5]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt4);
  AddSphereFaceGridTesselatedPoints(pt4, pt5);
  AddSphereFaceGridTesselatedPoints(pt5, pt1);
}

VOID ODodecahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();
  
  if( (GetType() == estODTile1) ||
      (GetType() == estODTile2) )
    {
      tpt2->SetHeightPartner(tpt4);
      tpt4->SetHeightPartner(tpt2);
    }
  else if( GetType() == estODTile3 )
    {
      tpt1->SetHeightPartner(tpt3);
      tpt3->SetHeightPartner(tpt1);
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID ODodecahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;
  
  m_listFaceTransformMatrix.RemoveAll();
  
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();

    /////////////////////////////////////////////////////////////////////////
    // Top

    // Face 1
    glPushMatrix();
    {
      // Identity
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glRotatef(120, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glRotatef(240, 0.0, GOLDENRATIO, INVGOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Top Side

    // Face 4
    glPushMatrix();
    {
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 5
    glPushMatrix();
    {
      glRotatef(120, 0.0, -GOLDENRATIO, -INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 6
    glPushMatrix();
    {
      glRotatef(240, 0.0, -GOLDENRATIO, -INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Bottom

    // Face 7
    glPushMatrix();
    {
      glRotatef(120, 0.0, -GOLDENRATIO, INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 8
    glPushMatrix();
    {
      glRotatef(120, 0.0, -GOLDENRATIO, -INVGOLDENRATIO);
      glRotatef(120, 0.0, -GOLDENRATIO,  INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 9
    glPushMatrix();
    {
      glRotatef(240, 0.0, -GOLDENRATIO, -INVGOLDENRATIO);
      glRotatef(120, 0.0, -GOLDENRATIO, INVGOLDENRATIO);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Bottom Side

    // Face 10
    glPushMatrix();
    {
      glRotatef(240, INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 11
    glPushMatrix();
    {
      glRotatef(120, -1.0, 1.0, 1.0);
      glRotatef(240, INVGOLDENRATIO, 0.0, GOLDENRATIO);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 12
    glPushMatrix();
    {
      glRotatef(240, -1.0, 1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
  }
  glPopMatrix();
}

VOID ODodecahedron::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;
  Vector v;

  m_listTileTransformMatrix.RemoveAll();

  v.Set(0.0, (GOLDENRATIO+2.0)/5.0, (INVGOLDENRATIO + 2*GOLDENRATIO + 2.0)/5.0 );

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(1*72));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(2*72));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 4
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(3*72));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 5
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(4*72));
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID ODodecahedron::FindMatrices(Matrix &/*m1*/, Matrix &/*m2*/, TilePoint */*tpt1*/)
{
  ASSERT(FALSE);
}

///////////////////////////////////////////////////////////////////////////////

VOID ODodecahedron::OutputES(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "ododecahedron\n\n" );

  m_pTile->OutputES(fpOutputFile);

  OutputESColor(fpOutputFile);
}
